import numpy as np
from typing import Union

from . import ActivationBaseClass


class Sigmoid(ActivationBaseClass):
    """Sigmoid function

    Class that holds the sigmoid function and gradient as discussed in `[1]`_

    Parameters
    ----------
    beta : int or float, optional, default=1
           Positive non-zero value that controls the steepness of the Sigmoid function.

    See also
    --------
    Identity, SoftPlus, Swish

    References
    ----------
    _`[1]` Villmann, T., Ravichandran, J., Villmann, A., Nebel, D., & Kaden, M. (2019). "Activation
    Functions for Generalized Learning Vector Quantization - A Performance Comparison", 2019.
    """

    __slots__ = "beta"

    def __init__(self, beta: Union[int, float] = 1):
        if beta <= 0:
            raise ValueError(
                "{}: Expected beta > 0, but got beta = {}".format(
                    type(self).__name__, beta
                )
            )

        self.beta = beta

    def __call__(self, x: np.ndarray) -> np.ndarray:
        r"""Computes the sigmoid function:
            .. math::

                f(\mathbf{x}) = \frac{1}{e^{-\beta \cdot \mathbf{x}} + 1}

        Parameters
        ----------
        x : ndarray of any shape.

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the sigmoid function.
        """
        return np.asarray(1.0 / (np.exp(-self.beta * x) + 1.0))

    def gradient(self, x: np.ndarray) -> np.ndarray:
        r"""Computes the sigmoid function's gradient with respect to x:
           .. math::

               \frac{\partial f}{\partial \mathbf{x}} = \frac{(\beta \cdot e^{\beta \cdot \mathbf{x})}}{(e^{\beta \cdot \mathbf{x}} + 1)^2}

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
           Elementwise evaluation of the sigmoid function's gradient.
        """
        exp = np.exp(self.beta * x)
        return np.asarray((self.beta * exp) / (exp + 1.0) ** 2)



from ._base import DiscriminantBaseClass
import numpy as np


class RelativeDistance(DiscriminantBaseClass):
    """Relative distance function

    Class that holds the relative distance function and gradient as described in `[1]`_.

    References
    ----------
    _`[1]` Sato, A., and Yamada, K. (1996) "Generalized Learning Vector Quantization."
    Advances in Neural Network Information Processing Systems, 423–429, 1996."""

    __slots__ = ()

    def __call__(self, dist_same: np.ndarray, dist_diff: np.ndarray) -> np.ndarray:
        r"""The relative distance discriminant function for a single sample (:math:`\mathbf{x}`):
            .. math::

                \mu(\mathbf{x}) = \frac{d(\mathbf{x}, \mathbf{w}_1) - d(\mathbf{x}, \mathbf{w}_0)}{d(
                \mathbf{x}, \mathbf{w}_1) + d(\mathbf{x}, \mathbf{w}_0)},

        with :math:`\mathbf{w}_1` the prototype with the same label and :math:`\mathbf{w}_0` the
        prototype with a different label.

        Parameters
        ----------
        dist_same : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with the same label.
        dist_diff : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with a different label.

        Returns
        -------
        ndarray with shape (n_samples, 1)
            Evaluation of the relative distance discriminative function.

        """
        return (dist_same - dist_diff) / (dist_same + dist_diff)

    def gradient(
        self, dist_same: np.ndarray, dist_diff: np.ndarray, same_label: bool
    ) -> np.ndarray:
        r"""Computes the relative distance discriminant function's gradient.

            1. The partial derivative with respect to the closest prototypes with the same label (same_label=True):

            .. math::

                \frac{\partial \mu}{\partial \mathbf{w}_1} = \frac{2 \cdot d(\mathbf{x},
                \mathbf{w}_0))}{(d(\mathbf{x}, \mathbf{w}_1) + d(\mathbf{x}, \mathbf{w}_0))^2}.

            2. The partial derivative with respect to the closest prototypes with a different label (same_label=False):

            .. math::

                \frac{\partial \mu}{\partial \mathbf{w}_0} = \frac{-2 \cdot d(\mathbf{x},
                \mathbf{w}_1))}{(d(\mathbf{x}, \mathbf{w}_1) + d(\mathbf{x}, \mathbf{w}_0))^2},

            with :math:`d(\mathbf{x}, \mathbf{w}_1)` the distance to the prototype with the same label
            and :math:`d(\mathbf{x}, \mathbf{w}_0)` the distance to the closest prototype with a
            different label.

        Parameters
        ----------
        dist_same : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with the same label.
        dist_diff : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with a different label.
        same_label : bool
            Indicating if the derivative with respect to a prototype with the same label (True) or
            a different label (False) needs to be calculated.

        Returns
        -------
        ndarray with shape (n_samples, 1)
            Evaluation of the relative distance function's gradient.

        """
        if same_label:
            return _gradient_same(dist_same, dist_diff)
        return _gradient_diff(dist_same, dist_diff)


def _gradient_same(dist_same: np.ndarray, dist_diff: np.ndarray) -> np.ndarray:
    return 2 * dist_diff / (dist_same + dist_diff) ** 2


def _gradient_diff(dist_same: np.ndarray, dist_diff: np.ndarray) -> np.ndarray:
    return -2 * dist_same / (dist_same + dist_diff) ** 2



from ._base import ActivationBaseClass
from ._identity import Identity
from ._sigmoid import Sigmoid
from ._soft_plus import SoftPlus
from ._swish import Swish

from sklvq._utils import _import_from_string

__all__ = ["ActivationBaseClass", "Identity", "Sigmoid", "SoftPlus", "Swish"]

ALIASES = {"soft+": "soft-plus"}


def import_from_string(class_string, valid_strings=None) -> type:
    return _import_from_string(
        __name__, class_string, ALIASES, "activation_type", valid_strings
    )



import numpy as np

from sklearn import datasets
from sklearn import preprocessing
from sklearn.model_selection import (
    GridSearchCV,
    RepeatedStratifiedKFold,
)
from sklearn.pipeline import make_pipeline

from .. import GLVQ


def test_shared_memory_glvq():
    X, y = datasets.load_iris(return_X_y=True)
    m = GLVQ(activation_type="identity").fit(X, y)

    p = m.prototypes_
    m.set_model_params(np.random.random(size=(3, 4)))
    assert np.shares_memory(p, m.get_variables())
    assert np.all(m.get_variables() == m.prototypes_.ravel())

    model_params = m.to_model_params_view(m.get_variables())
    assert np.all(m.prototypes_.shape == model_params.shape)
    assert np.shares_memory(m.prototypes_, m.get_variables())
    assert np.shares_memory(model_params, m.get_variables())


def test_glvq():
    iris = datasets.load_iris()

    estimator = GLVQ(random_state=31415)
    pipeline = make_pipeline(preprocessing.StandardScaler(), estimator)

    scipy_solvers_types = ["lbfgs", "bfgs"]
    # Run each solver ones
    solvers_types = [
        "steepest-gradient-descent",
        "waypoint-gradient-descent",
        "adaptive-moment-estimation",
    ]
    discriminant_types = ["relative-distance"]

    # Every compatible distance
    distance_types = ["squared-euclidean", "euclidean"]

    # Every compatible activation
    activation_types = ["identity", "sigmoid", "soft-plus", "swish"]

    param_grid = [
        {
            "glvq__solver_type": scipy_solvers_types,
            "glvq__solver_params": [{"jac": None}, {}],
            "glvq__discriminant_type": discriminant_types,
            "glvq__distance_type": distance_types,
            "glvq__activation_type": activation_types,
        },
        {
            "glvq__solver_type": solvers_types,
            "glvq__discriminant_type": discriminant_types,
            "glvq__distance_type": distance_types,
            "glvq__activation_type": activation_types,
        },
    ]

    repeated_kfolds = RepeatedStratifiedKFold(n_splits=2, n_repeats=1)

    search = GridSearchCV(
        pipeline,
        param_grid,
        scoring=["accuracy", "roc_auc_ovo", "precision_macro", "recall_macro"],
        cv=repeated_kfolds,
        return_train_score=True,
        refit="roc_auc_ovo",
    )

    search.fit(iris.data, iris.target)

    assert np.all(search.cv_results_["mean_train_roc_auc_ovo"] > 0.75)
    assert np.all(search.cv_results_["mean_test_roc_auc_ovo"] > 0.75)

    print("\nBest parameter (CV roc_auc=%0.3f):" % search.best_score_)
    print(search.best_params_)



from ._base import DiscriminantBaseClass
from ._relative_distance import RelativeDistance
from sklvq._utils import _import_from_string

__all__ = ["DiscriminantBaseClass", "RelativeDistance"]

ALIASES = {}


def import_from_string(class_string, valid_strings=None) -> type:
    return _import_from_string(
        __name__, class_string, ALIASES, "discriminant_type", valid_strings
    )



from . import DistanceBaseClass

import numpy as np
from scipy.spatial.distance import cdist

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import GMLVQ


class AdaptiveSquaredEuclidean(DistanceBaseClass):
    """Adaptive squared Euclidean distance

    Class that holds the adaptive squared Euclidean distance function and its gradient as
    described in `[1]`_ and `[2]`_.

    Parameters
    ----------
    force_all_finite  : {True, False, "allow-nan"}
        Parameter to indicate that NaNLVQ distance variant should be used. If true no nans are
        allowed. If False or "allow-nan", nans are allowed.

    See also
    --------
    Euclidean, SquaredEuclidean, LocalAdaptiveSquaredEuclidean

    Notes
    -----
    Compatible with the :class:`.GMLVQ` algorithm (only).

    References
    ----------
    _`[1]` Schneider, P. (2010). Advanced methods for prototype-based classification. Groningen.

    _`[2]` Schneider, P., Biehl, M., & Hammer, B. (2009). Adaptive Relevance Matrices in Learning
    Vector Quantization. Neural Computation, 21(12), 3532–3561."""

    __slots__ = ()

    def __call__(self, data: np.ndarray, model: "GMLVQ") -> np.ndarray:
        r"""Computes the adaptive squared Euclidean distance:

            .. math::
                d^{\Lambda}(\mathbf{w}, \mathbf{x}) = (\mathbf{x} - \mathbf{w})^{\top}
                \Lambda (\mathbf{x} - \mathbf{w})

            with the relevance matrix :math:`\Lambda = \Omega^{\top} \Omega`, the  prototype
            :math:`\mathbf{w}`, and sample :math:`\mathbf{x}`.

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distance gradient to the prototypes of the model need to be
            computed.
        model : GMLVQ
            The model instance, containing the prototypes and relevance matrix.

        Returns
        -------
        ndarray with shape (n_samples, n_prototypes)
           Evaluation of the distance between each sample in the data and prototype of the model.
        """

        if model.force_all_finite == "allow-nan":
            # RM because VI is filtered out of the  kwargs by cdist...
            return cdist(
                data,
                model.prototypes_,
                _nan_mahalanobis,
                RM=model._compute_lambda(model.omega_),
            )

        return (
            cdist(
                data,
                model.prototypes_,
                "mahalanobis",
                VI=model._compute_lambda(model.omega_),
            )
            ** 2
        )

    def gradient(
        self, data: np.ndarray, model: "GMLVQ", i_prototype: int
    ) -> np.ndarray:
        r"""Computes the gradient of the adaptive squared euclidean distance function,
        with respect to a single prototype:

            .. math::
                \frac{\partial d}{\partial \mathbf{w_i}} = -2 \Lambda (\mathbf{x} -
                \mathbf{w_i}),

        and the omega matrix (per element):

            .. math::
                \frac{\partial d}{\partial \Omega_{lm}} =  2 \sum_i (x^i - w^i)
                \Omega_{li} (x^m - w^m)

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distance gradient to the prototypes of the model need to be
            computed.
        model : GMLVQ
            The model instance, containing the prototypes and relevance matrix.
        i_prototype : int
            An integer index value of the relevant prototype

        Returns
        -------
        ndarray with shape (n_samples, n_features + n_omega_elements)
            The gradient of the prototype and omega matrix with respect to each data sample.
        """

        (prototypes, omega) = model.get_model_params()

        prototype = prototypes[i_prototype, :]

        (num_samples, _) = data.shape

        distance_gradient = np.empty(
            (num_samples, prototype.size + omega.size), dtype="float64", order="C"
        )

        difference = data - prototype

        if model.force_all_finite == "allow-nan":
            difference[np.isnan(difference)] = 0.0

        # Prototype gradient directly computed in the distance_gradient
        _prototype_gradient(
            difference, omega, out=distance_gradient[:, : prototype.size]
        )

        # Omega view created in to hold the shape the _omega_gradient functions needs to output in.
        distance_gradient_omega_view = distance_gradient[:, prototype.size :].reshape(
            (num_samples, *omega.shape)
        )
        # Omega gradient indirectly computed in the distance_gradient via the view of the memory
        # in a different shape.
        _omega_gradient(difference, omega, out=distance_gradient_omega_view)

        return distance_gradient


def _nan_mahalanobis(sample, prototype, RM=None):
    # The NaNLVQ variant of the mahalanobis distance
    difference = sample - prototype
    difference[np.isnan(difference)] = 0.0
    return difference.dot(RM).dot(difference.T)
    # return np.einsum("i, ij, j ->", difference, relevance_matrix, difference)


def _prototype_gradient(
    difference: np.ndarray, omega: np.ndarray, out=None
) -> np.ndarray:
    # The gradient with respect to a prototype. Equivalent to np.dot(-2.0 * difference.dot(
    # omega.T.dot(omega)).
    return np.einsum("ji,ik ->jk", -2.0 * difference, np.dot(omega.T, omega), out=out)


def _omega_gradient(difference: np.ndarray, omega: np.ndarray, out=None) -> np.ndarray:
    # The gradient with respect to the omega matrix.
    return np.einsum(
        "ij,jk->jik", np.dot(omega, difference.T), (2.0 * difference), out=out
    )



import numpy as np
from sklearn.utils import shuffle

from . import SolverBaseClass
from ..objectives import ObjectiveBaseClass
from ._base import _update_state

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from sklvq.models import LVQBaseClass

STATE_KEYS = ["variables", "nit", "fun", "nfun", "tfun", "step_size"]


class WaypointGradientDescent(SolverBaseClass):
    r""" Waypoint gradient descent (WGD)

    Implements the waypoint average optimization algorithm `[1]`_. Implementation and description is
    inspired by `[2]`_.

    The algorithm keeps a rolling average of the last ``k`` model parameters. After ``k`` steps the
    algorithms will compare the cost of the average model parameters (:math:`\hat{
    \mathbf{\theta}}`) versus a "regular" update of the model parameters (:math:`\tilde{\mathbf{
    \theta}}`).

    .. math::
        \tilde{\mathbf{\theta}} &= \theta_t - \eta \cdot \frac{\nabla E(\theta_t)}{||\nabla E(\theta_t)||}  \\
        \hat{\mathbf{\theta}} &= \frac{1}{k} \sum_{i=0}^{k - 1} \mathbf{\theta}_{t_i}

    If the regular step results in a lower cost (:math:`E(\tilde{\mathbf{\theta}}) < E(\hat{
    \mathbf{\theta}})`), the ``step_size`` is increased by multiplying with the ``gain`` factor:

    .. math::
        \mathbf{\theta}_{t+1} &= \tilde{\mathbf{\theta}}  \\
        \eta &= gain \cdot \eta.

    If the average step results in a lower cost ((:math:`E(\hat{\mathbf{\theta}}) < E(\tilde{
    \mathbf{\theta}})`) the ``step_size`` is decreased by multiplying with
    the ``loss`` factor:

    .. math::
        \mathbf{\theta}_{t+1} &= \hat{\mathbf{\theta}} \\
        \eta &= loss \cdot \eta.

    Note that the solver uses the normalized objective gradient to update the model.

    Parameters
    ----------
    objective: ObjectiveBaseClass, required
       This is set by the algorithm. See :class:`sklvq.models.GLVQ`, :class:`sklvq.models.GMLVQ`,
        and :class:`sklvq.models.LGMLVQ`.

    max_runs: int
        Maximum number of runs/epochs that will be computed. Should be >= k. Early stopping can
        be implemented by providing a ``callback`` function that returns True when the solver should
        stop.

    step_size: float or ndarray
        The step size to control the learning rate of the model parameters. If the same step size
        should be used for all parameters (e.g., prototypes and omega) then a single float is
        sufficient. If separate initial step_sizes should be used per model parameter then this
        should be specified by using a ndarray.

        Whenever the averge update is accepted (has a lower cost) the step sizes are multiplied
        with the ``loss`` factor. When the "regular" update is accepted then the step size(s) are
        multiplied by the ``gain`` factor.

    loss: float
        Should be a value less than 1. Controls the step size change factor when an
        average waypoint step is accepted.
    gain: float
        Should be a value greater than 1. Controls the step size change factor when a
        regular update step is accepted.
    k: int
        The number of runs used to compute the average waypoint over.

    callback: callable
        Callable with signature callable(state). If the callable returns True the solver
        will stop even if max_runs is not reached yet. The state object contains the following:

        - "variables"
            Concatenated 1D ndarray of the model's parameters
        - "nit"
            The current iteration counter.
        - "fun"
            The accepted cost.
        - "nfun"
            The cost of the regular update step.
        - "tfun"
            The cost of the "tentative" update, i.e., the average of the past k updates.
        - "step_size"
            The current step_size(s)


    References
    ----------
    _`[1]` Papari, G., and Bunte, K., and Biehl, M. (2011) "Waypoint averaging and step size
    control in learning by gradient descent" Mittweida Workshop on Computational
    Intelligence (MIWOCI) 2011.

    _`[2]` LeKander, M., Biehl, M., & De Vries, H. (2017). "Empirical evaluation of gradient
    methods for matrix learning vector quantization." 12th International Workshop on
    Self-Organizing Maps and Learning Vector Quantization, Clustering and Data
    Visualization, WSOM 2017.
    """

    def __init__(
        self,
        objective: ObjectiveBaseClass,
        max_runs: int = 10,
        step_size: Union[float, np.ndarray] = 0.1,
        loss: float = 2 / 3,
        gain: float = 1.1,
        k: int = 3,
        callback: callable = None,
    ):
        super().__init__(objective)
        if max_runs <= k or max_runs <= 0:
            raise ValueError(
                "{}:  Expected 0 < max_runs > k, but got max_runs = {}".format(
                    type(self).__name__, max_runs
                )
            )
        self.max_runs = max_runs

        if np.any(step_size <= 0):
            raise ValueError(
                "{}:  Expected step_size to be > 0, but got step_size = {}".format(
                    type(self).__name__, step_size
                )
            )
        self.step_size = step_size

        if loss <= 0 or loss > 1:
            raise ValueError(
                "{}: Expected loss to be > 0 and < 1, but got loss = {}".format(
                    type(self).__name__, loss
                )
            )
        self.loss = loss

        if gain < 1:
            raise ValueError(
                "{}: Expected gain to be >= 1, but got gain = {}".format(
                    type(self).__name__, gain
                )
            )
        self.gain = gain

        if k <= 1:
            raise ValueError(
                "{}: Expected k to be >= 2, but got k = {}".format(
                    type(self).__name__, k
                )
            )
        self.k = k

        if callback is not None:
            if not callable(callback):
                raise ValueError(
                    "{}:  callback is not callable.".format(type(self).__name__)
                )
        self.callback = callback

    def solve(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        model: "LVQBaseClass",
    ):
        """Solve function that gets called by the fit method of the models.

        Performs the steps of the waypoint gradient descent optimization method.

        Parameters
        ----------
        data : ndarray of shape (n_samples, n_features)
            The data.

        labels : ndarray of size (n_samples)
            The labels of the samples in the data.

        model : LVQBaseClass
            The initial model that will also hold the final result
        """

        previous_waypoints = np.empty((self.k, model.get_variables().size))
        tentative_model_variables = np.empty(model.get_variables().size)

        step_size = self.step_size

        if self.callback is not None:
            variables = np.copy(model.get_variables())
            cost = self.objective(model, data, labels)
            state = _update_state(
                STATE_KEYS, variables=variables, nit="Initial", nfun=cost, fun=cost
            )
            if self.callback(state):
                return

        # Initial runs to get enough gradients to average.
        for i_run in range(0, self.k):
            shuffled_indices = shuffle(
                range(0, labels.size), random_state=model.random_state_
            )

            shuffled_data = data[shuffled_indices, :]
            shuffled_labels = labels[shuffled_indices]

            objective_gradient = self.objective.gradient(
                model, shuffled_data, shuffled_labels
            )

            # Normalize the gradient by gradient/norm(gradient)
            model.normalize_variables(objective_gradient)

            # Multiply params by step_size
            model.mul_step_size(step_size, objective_gradient)

            model.set_variables(
                np.subtract(  # returns out=objective_gradient
                    model.get_variables(),
                    objective_gradient,
                    out=objective_gradient,
                )
            )

            previous_waypoints[np.mod(i_run, self.k), :] = model.get_variables()

            if self.callback is not None:
                cost = self.objective(model, data, labels)
                state = _update_state(
                    STATE_KEYS,
                    variables=np.copy(model.get_variables()),
                    nit=i_run + 1,
                    nfun=cost,
                    fun=cost,
                    step_size=step_size,
                )
                if self.callback(state):
                    return

        # The remainder of the runs
        for i_run in range(self.k, self.max_runs):
            shuffled_indices = shuffle(
                range(0, labels.size), random_state=model.random_state_
            )

            shuffled_data = data[shuffled_indices, :]
            shuffled_labels = labels[shuffled_indices]

            objective_gradient = self.objective.gradient(
                model, shuffled_data, shuffled_labels
            )

            # Normalize the gradient by gradient/norm(gradient)
            model.normalize_variables(objective_gradient)

            # Multiply params by step_size
            model.mul_step_size(step_size, objective_gradient)

            new_model_variables = np.subtract(  # returns out=objective_gradient
                model.get_variables(),
                objective_gradient,
                out=objective_gradient,
            )

            # Tentative average update
            np.mean(previous_waypoints, axis=0, out=tentative_model_variables)

            # Update model
            model.set_variables(tentative_model_variables)

            # Compute cost of tentative update step
            tentative_cost = self.objective(model, shuffled_data, shuffled_labels)

            # Update model
            model.set_variables(new_model_variables)

            # Compute cost of regular update step
            new_cost = self.objective(model, shuffled_data, shuffled_labels)

            if tentative_cost < new_cost:
                model.set_variables(tentative_model_variables)
                step_size = self.loss * step_size
                accepted_cost = tentative_cost
            else:
                step_size = self.gain * step_size
                accepted_cost = new_cost

            # Administration. Store the models parameters.
            previous_waypoints[np.mod(i_run, self.k), :] = model.get_variables()

            if self.callback is not None:
                state = _update_state(
                    STATE_KEYS,
                    variables=np.copy(model.get_variables()),
                    nit=i_run + 1,
                    tfun=tentative_cost,
                    nfun=new_cost,
                    fun=accepted_cost,
                    step_size=step_size,
                )
                if self.callback(state):
                    return



from typing import Tuple
from typing import Union

import numpy as np
from sklearn.utils.validation import check_is_fitted, check_array

from . import LVQBaseClass
from ..objectives import GeneralizedLearningObjective

ModelParamsType = Tuple[np.ndarray, np.ndarray]

DISTANCES = [
    "adaptive-squared-euclidean",
]

SOLVERS = [
    "adaptive-moment-estimation",
    "broyden-fletcher-goldfarb-shanno",
    "limited-memory-bfgs",
    "steepest-gradient-descent",
    "waypoint-gradient-descent",
]


class GMLVQ(LVQBaseClass):
    r"""Generalized Matrix Learning Vector Quantization

    This model uses the :class:`sklvq.objectives.GeneralizedLearningObjective` as its objective
    function `[1]`_. In addition to learning the positions of the prototypes it learns a relevance
    matrix that is used in the distance functions `[2]`_.

    Parameters
    ----------
    distance_type : {"adaptive-squared-euclidean"} or Class, default="squared-euclidean"
        Distance function that employs a relevance matrix in its calculation.

        - "adaptive-squared-euclidean"
            See :class:`sklvq.distances.AdaptiveSquaredEuclidean`

    distance_params : Dict, default=None
        Parameters passed to init of distance callable

    activation_type : {"identity", "sigmoid", "soft+", "swish"} or Class, default="sigmoid"
        Parameters passed to init of activation function. See the documentation of the activation
        functions for parameters and defaults.

        - "identity"
            See :class:`sklvq.activations.Identity`
        - "sigmoid"
            See :class:`sklvq.activations.Sigmoid`
        - "soft+"
            See :class:`sklvq.activations.SoftPlus`
        - "swish"
            See :class:`sklvq.activations.Swish`

    activation_params : Dict, default=None
        Parameters passed to init of activation function. See the documentation of activation
        functions for function dependent parameters and defaults.

    discriminant_type : {"relative-distance"} or Class, default = "relative-distance"
        The discriminant function.  Note that different discriminant type may require to rewrite
        the ``decision_function`` and ``predict_proba`` methods.

        - "relative-distance"
            See :class:`sklvq.discriminants.RelativeDistance`

    discriminant_params : Dict, default=None
        Parameters passed to init of discriminant callable. See the documentation of the
        discriminant functions for parameters and defaults.

    solver_type : {"sgd", "wgd", "adam", "lbfgs", "bfgs"},
        The solver used for optimization

        - "sgd" or "steepest-gradient-descent"
            See :class:`sklvq.solvers.SteepestGradientDescent`.
        - "wgd" or "waypoint-gradient-descent"
            See :class:`sklvq.solvers.WaypointGradientDescent`.
        - "adam" or "adaptive-moment-estimation"
            See :class:`sklvq.solvers.AdaptiveMomentEstimation`.
        - "bfgs" or "broyden-fletcher-goldfarb-shanno"
            Implementation from scipy package.
        - "lbfgs" or "limited-memory-bfgs"
            Implementation from scipy package.

    solver_params : dict, default=None
        Parameters passed to init of solvers. See the documentation of the solvers relevant
        parameters and defaults.

    prototype_init: "class-conditional-mean" or ndarray, default="class-conditional-mean"
        Default will initiate the prototypes to the class conditional mean with a small random
        offset. Custom numpy array can be passed to change the initial positions of the prototypes.


    prototype_n_per_class: int or np.ndarray, optional, default=1
        Default will generate single prototype per class. In the case of unequal number of
        prototypes per class is needed, provide this as np.ndarray. For example,
        prototype_n_per_class = np.array([1, 6, 3]) this will result in one prototype for the first class,
        six for the second, and three for the third. Note that the order needs to be the same as the on in the
        classes\_ attribute, which is equal to calling np.unique(labels).

    relevance_init : {"identity", "random"} or np.ndarray, default="identity"
        Default will initiate the omega matrices to be the identity matrix. The rank of the matrix can be reduced by
        setting the ``relevance_n_components`` attribute `[3]`_.

    relevance_normalization: bool, optional, default=True
        Flag to indicate whether to normalize omega, whenever it is updated, such that the trace of the relevance matrix
        is equal to 1.

    relevance_n_components: str {"all"} or int, optional, default="all"
        For a square relevance matrix use the string "all" (default). For a rectangular relevance matrix use set the
        number of components explicitly by providing it as an int.

    random_state : int, RandomState instance, default=None
        Set the random number generation for reproducibility purposes. Used in random offset of prototypes and
        shuffling of the data in the solvers. Potentially, also used in the random generation of relevance matrix.

    force_all_finite : {True, "allow-nan"}, default=True
        Whether to raise an error on np.inf, np.nan, pd.NA in array. The possibilities are:

        - True: Force all values of array to be finite.
        - "allow-nan": accepts only np.nan and pd.NA values in array. Values cannot be infinite.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        Class labels for each output.

    prototypes_ : ndarray of shape (n_protoypes, n_features)
        Positions of the prototypes after ``fit(X, labels)`` has been called.

    prototypes_labels_ : ndarray of shape (n_prototypes)
        Labels for each prototypes. Labels are indexes to ``classes_``

    omega_: ndarray with size depending on initialization, default (n_features, n_features)
        Omega matrix that was found during training and defines the relevance matrix ``lambda_``.

    lambda_: ndarray of size (n_features, n_features)
        The relevance matrix ``omega_.T.dot(omega_)``

    omega_hat_: ndarray
        The omega matrix found by the eigenvalue decomposition of the relevance matrix ``lambda_``.
        The eigenvectors (columns of ``omega_hat_``) can be used to transform the X `[3]`_.

    eigenvalues_: ndarray
        The corresponding eigenvalues to ``omega_hat_`` found by the eigenvalue decomposition of
        the relevance matrix ``lambda_``

    References
    ----------
    _`[1]` Sato, A., and Yamada, K. (1996) "Generalized Learning Vector Quantization."
    Advances in Neural Network Information Processing Systems, 423–429, 1996.

    _`[2]` Schneider, P., Biehl, M., & Hammer, B. (2009). "Adaptive Relevance Matrices in
    Learning Vector Quantization" Neural Computation, 21(12), 3532–3561, 2009.

    _`[3]` Bunte, K., Schneider, P., Hammer, B., Schleif, F.-M., Villmann, T., & Biehl, M. (2012).
    "Limited Rank Matrix Learning, discriminative dimension reduction and visualization." Neural
    Networks, 26, 159–173, 2012."""

    classes_: np.ndarray
    prototypes_: np.ndarray
    prototypes_labels_: np.ndarray
    omega_: np.ndarray
    lambda_: np.ndarray
    omega_hat_: np.ndarray
    eigenvalues_: np.ndarray

    def __init__(
        self,
        distance_type: Union[str, type] = "adaptive-squared-euclidean",
        distance_params: dict = None,
        activation_type: Union[str, type] = "sigmoid",
        activation_params: dict = None,
        discriminant_type: Union[str, type] = "relative-distance",
        discriminant_params: dict = None,
        solver_type: Union[str, type] = "steepest-gradient-descent",
        solver_params: dict = None,
        prototype_init: Union[str, np.ndarray] = "class-conditional-mean",
        prototype_n_per_class: Union[int, np.ndarray] = 1,
        relevance_init="identity",
        relevance_normalization: bool = True,
        relevance_n_components: Union[str, int] = "all",
        relevance_regularization: Union[int, float] = 0,
        random_state: Union[int, np.random.RandomState] = None,
        force_all_finite: Union[str, bool] = True,
    ):
        self.activation_type = activation_type
        self.activation_params = activation_params
        self.discriminant_type = discriminant_type
        self.discriminant_params = discriminant_params
        self.relevance_init = relevance_init
        self.relevance_normalization = relevance_normalization
        self.relevance_n_components = relevance_n_components
        self.relevance_regularization = relevance_regularization

        super(GMLVQ, self).__init__(
            distance_type,
            distance_params,
            DISTANCES,
            solver_type,
            solver_params,
            SOLVERS,
            prototype_init,
            prototype_n_per_class,
            random_state,
            force_all_finite,
        )

    ###########################################################################################
    # The "Getter" and "Setter" that are used by the solvers to set and get model params.
    ###########################################################################################

    def set_variables(self, new_variables: np.ndarray) -> None:
        """

        Modifies the ``self._variables`` by copying the values of ``new_variables`` into the
        memory of ``self._variables``.

        Parameters
        ----------
        new_variables : ndarray
            1d numpy array that contains all the model parameters in continuous memory
        """
        np.copyto(self._variables, new_variables)
        if self.relevance_normalization:
            GMLVQ._normalise_omega(self.omega_)

    def set_model_params(self, new_model_params: ModelParamsType):
        """
        Changes the model's internal parameters. Copies the values of model_params into
        ``self.prototypes_`` and ``self.omega_`` therefor updating the ``self.variables_``
        array.

        Also normalized the relevance matrix if necessary.

        Parameters
        ----------
        new_model_params : tuple of ndarrays
            Shapes depend on  initialization but in the case of a square relevance matrix:
            tuple((n_prototypes, n_features), (n_features, n_features))
        """
        new_prototypes, new_omega = new_model_params

        self.set_prototypes(new_prototypes)
        self.set_omega(new_omega)

        if self.relevance_normalization:
            GMLVQ._normalise_omega(self.omega_)

    def get_model_params(self) -> ModelParamsType:
        """
        Returns a tuple of all model parameters. In this case the prototypes and omega matrix.

        Returns
        -------
        ndarray
             Returns a tuple of views, i.e., the prototypes and omega matrix.
        """
        return self.prototypes_, self.omega_

    ###########################################################################################
    # Specific "getters" and "setters" for GMLVQ
    ###########################################################################################

    def get_omega(self):
        """
        Convenience function to return ``self.omega_``

        Returns
        -------
            ndarray, with shape depending on initialization of omega.
        """
        return self.omega_

    def set_omega(self, omega):
        """
        Convenience function that makes sure to copy the value to ``self.omega_`` and not overwrite
        it.

        Parameters
        ----------
        omega : ndarray with same shape as ``self.omega_``
        """
        np.copyto(self.omega_, omega)

    ###########################################################################################
    # Functions to transform the 1D variables array to model parameters and back
    ###########################################################################################

    def to_model_params_view(self, var_buffer: np.ndarray) -> ModelParamsType:
        """

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        tuple
            Returns a tuple with the prototypes and omega matrix as ndarrays.
        """
        return (
            self.to_prototypes_view(var_buffer),
            self.to_omega(var_buffer),
        )

    def to_prototypes_view(self, var_buffer: np.ndarray) -> np.ndarray:
        """
        Returns a view (of the shape of the model's prototypes) into the provided variables
        buffer of the same size as the model's variables array.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray of shape (n_prototypes, n_features)
            Prototype view into the var_buffer.
        """
        return var_buffer[: self._prototypes_size].reshape(self._prototypes_shape)

    def to_omega(self, var_buffer: np.ndarray) -> np.ndarray:
        """
        Returns a view (of the shape of the model's omega) into the provided variables
        buffer of the same size as the model's variables array.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray
            Shape depending on initialization but in case of a square matrix (n_features,
            n_features.
        """
        return var_buffer[self._prototypes_size :].reshape(self._relevances_shape)

    ###########################################################################################
    # Solver Normalization functions
    ###########################################################################################

    def normalize_variables(self, var_buffer: np.ndarray) -> None:
        """
        Modifies the var_buffer as if it was the variables array provided
        by ``get_variables()``. Will select, reshape and normalize the correct parts of the
        variable buffer.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.
        """
        (prototypes, omega) = self.to_model_params_view(var_buffer)

        self._normalize_prototypes(prototypes)
        self._normalise_omega(omega)

    @staticmethod
    def _normalise_omega(omega: np.ndarray) -> None:
        np.divide(omega, np.sqrt(np.einsum("ji, ji", omega, omega)), out=omega)

    ###########################################################################################
    # Solver helper functions
    ###########################################################################################

    def add_partial_gradient(self, gradient, partial_gradient, i_prototype) -> None:
        """
        Adds the partial gradient to the correct part of the gradient, which  depends on
        ``i_prototype``.

        Parameters
        ----------
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.

        partial_gradient : ndarray
            1d array containing the partial gradient.

        i_prototype : int
            The index of the prototype to which the partial gradient was  computed.
        """
        n_features = self.n_features_in_

        prots_view = self.to_prototypes_view(gradient)
        np.add(
            prots_view[i_prototype, :],
            partial_gradient[:n_features],
            out=prots_view[i_prototype, :],
        )

        omega_view = gradient[self.prototypes_.size :]
        np.add(omega_view, partial_gradient[n_features:], out=omega_view)

    def mul_step_size(
        self, step_sizes: Union[int, float, np.ndarray], gradient: np.ndarray
    ) -> None:
        """
        If step sizes is a scalar value just multiplies the gradient with the step size. If it
        is an array (with same length as number of model parameters) each model parameter is
        multiplied by its own step size.

        Parameters
        ----------
        step_sizes : float or ndarray
            The scalar or list of values containing the step sizes.
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.
        """
        if isinstance(step_sizes, int) | isinstance(step_sizes, float):
            gradient *= step_sizes
            return

        # else it's a np.ndarray...
        if isinstance(step_sizes, np.ndarray):
            if step_sizes.size > 1:
                prototypes, omega = self.to_model_params_view(gradient)
                prototypes *= step_sizes[0]
                omega *= step_sizes[1]
            # if size is more than 2 it just ignores the additional values.

    ###########################################################################################
    # Initialization functions
    ###########################################################################################

    def _init_variables(self) -> None:
        self._variables = np.empty(
            self._prototypes_size + self._relevances_size,
            dtype="float64",
            order="C",
        )

    def _check_model_params(self):
        self._check_prototype_params()
        self._check_relevances_params()

    def _init_model_params(self, X, y) -> None:
        self._init_prototypes(X, y)
        self._init_relevances()

    def _check_relevances_params(self):
        relevance_normalization = self.relevance_normalization
        if not isinstance(relevance_normalization, bool):
            raise ValueError("Provided normalization is invalid.")

        relevance_n_components = self.relevance_n_components
        if isinstance(relevance_n_components, str):
            if relevance_n_components == "all":
                self._relevances_shape = (self.n_features_in_, self.n_features_in_)
            else:
                raise ValueError("Provided n_components is invalid.")
        elif isinstance(relevance_n_components, int):
            if (
                self.relevance_n_components >= 1
                and relevance_n_components <= self.n_features_in_
            ):
                self._relevances_shape = (relevance_n_components, self.n_features_in_)
            else:
                raise ValueError("Provided n_components is invalid.")
        else:
            raise ValueError("Provided n_components is invalid.")

        self._relevances_size = np.prod(self._relevances_shape)

    def _init_relevances(self):
        self.omega_ = self.to_omega(self._variables)

        if isinstance(self.relevance_init, str):
            if self.relevance_init == "identity":
                self.set_omega(np.eye(*self._relevances_shape))
            elif self.relevance_init == "random":
                self.set_omega(
                    self.random_state_.uniform(
                        low=0, high=1, size=self._relevances_shape
                    )
                )
            else:
                raise ValueError("Provided relevance_init is invalid.")
        else:
            raise ValueError("Provided relevance_init is invalid.")

        if self.relevance_normalization:
            GMLVQ._normalise_omega(self.omega_)

    def _init_objective(self):
        self._objective = GeneralizedLearningObjective(
            self.activation_type,
            self.activation_params,
            self.discriminant_type,
            self.discriminant_params,
        )

    ###########################################################################################
    # Other Algorithm specific stuff.
    ###########################################################################################

    def _after_fit(self, X: np.ndarray, y: np.ndarray):
        self.lambda_ = GMLVQ._compute_lambda(self.omega_)

        # Eigenvalues and column eigenvectors return in ascending order
        eigenvalues, omega_hat = np.linalg.eigh(self.lambda_)

        # Flip (reverse the order to descending) before assigning.
        self.eigenvalues_ = np.flip(eigenvalues)

        # eigenvectors are column matrix in ascending order. Flip the columns and transpose the matrix
        # to get the descending ordered row matrix.
        self.omega_hat_ = np.flip(omega_hat, axis=1).T

    @staticmethod
    def _compute_lambda(omega):
        # Equivalent to omega.T.dot(omega), but faster (?)
        return np.einsum("ji, jk -> ik", omega, omega)

    ###########################################################################################
    # Transformer related functions
    ###########################################################################################

    def fit_transform(
        self, data: np.ndarray, y: np.ndarray, **transform_params
    ) -> np.ndarray:
        r"""
        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            Data used for fit and that will be transformed.
        y : np.ndarray with length (n_samples)
            Labels corresponding to the X samples.
        transform_params :
            Parameters passed to transform function

        Returns
        -------
        The data projected on columns of ``omega_hat_`` with shape (n_samples, n_columns)
        """
        return self.fit(data, y).transform(data, **transform_params)

    def transform(self, X: np.ndarray, scale: bool = False) -> np.ndarray:
        r"""
        Parameters
        ----------
        X : np.ndarray with shape (n_samples, n_features)
            Data that needs to be transformed
        scale : {True, False}, default = False
            Controls if the eigenvectors the data is projected on are scaled by the square root
            of their eigenvalues.

        Returns
        -------
        The data projected on columns of ``omega_hat_`` with shape (n_samples, n_columns)
        """
        X = check_array(X)

        check_is_fitted(self)

        transformation_matrix = self.omega_hat_
        if scale:
            transformation_matrix = (
                np.sqrt(np.absolute(self.eigenvalues_[:, None])) * transformation_matrix
            )

        data_new = X.dot(transformation_matrix.T)

        return data_new



import numpy as np
import pytest

from .._steepest_gradient_descent import SteepestGradientDescent
from sklvq.objectives import GeneralizedLearningObjective
from sklvq.models import GLVQ
from sklearn.datasets import load_iris


def test_steepest_gradient_descent():

    objective = GeneralizedLearningObjective(
        activation_type="identity",
        activation_params=None,
        discriminant_type="relative-distance",
        discriminant_params=None,
    )

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, max_runs=-1)

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, max_runs=0)

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, batch_size=-1)

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, step_size=-1)

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, step_size=np.array([1, -1]))

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, step_size=np.array([1, 0]))

    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, callback=0)

    model = GLVQ()
    X, y = load_iris(return_X_y=True)
    with pytest.raises(ValueError):
        SteepestGradientDescent(objective, batch_size=100000).solve(X, y, model)



from abc import ABC, abstractmethod

import numpy as np


class ActivationBaseClass(ABC):
    """Activation base class

    Abstract class for implementing activation functions. Provides abstract methods with
    expected call signatures.

    Custom activation function '__init__' should accept any parameters as key-value pairs.

    See also
    --------
    Identity, Sigmoid, SoftPlus, Swish
    """

    __slots__ = ()

    @abstractmethod
    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Should implement an activation function

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
            Should perform an elementwise evaluation of some activation function.

        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def gradient(self, x: np.ndarray) -> np.ndarray:
        """
        Should implement the activation function's  gradient

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
            Should return the elementwise evaluation of the activation function's gradient.

        """
        raise NotImplementedError("You should implement this!")



from . import ActivationBaseClass
import numpy as np

from typing import Union


class Swish(ActivationBaseClass):
    """Swish function

    Class that holds the swish function and gradient as discussed in `[1]`_

    Parameters
    ----------
    beta : int, float, default=1
          Positive non-zero value that controls the steepness of the Swish function.

    See also
    --------
    Identity, Sigmoid, SoftPlus

    References
    ----------
    _`[1]` Villmann, T., Ravichandran, J., Villmann, A., Nebel, D., & Kaden, M. (2019). "Activation
    Functions for Generalized Learning Vector Quantization - A Performance Comparison", 2019.
    """

    __slots__ = "beta"

    def __init__(self, beta: Union[int, float] = 1):
        if beta <= 0:
            raise ValueError(
                "The activation function {} expects beta > 0, but got beta = {}".format(
                    type(self).__name__, beta
                )
            )

        self.beta = beta

    def __call__(self, x: np.ndarray) -> np.ndarray:
        r"""Implements the swish function:
            .. math::

                f(\mathbf{x}) = \frac{\mathbf{x}}{1 + e^{-\beta \cdot \mathbf{x}}}

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the swish function.
        """
        return _swish(x, self.beta)

    def gradient(self, x: np.ndarray) -> np.ndarray:
        r"""Implements the sigmoid function's gradient:
            .. math::

                \frac{\partial f}{\partial \mathbf{x}} = \beta \cdot f(\mathbf{x}) + (\frac{1}{1 + e^{-\beta \cdot \mathbf{x}}}) \cdot (1 - \beta \cdot f(\mathbf{x}))

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the swish function's gradient.
        """
        return (self.beta * _swish(x, self.beta)) + (
            _sgd(x, self.beta) * (1 - self.beta * _swish(x, self.beta))
        )


def _sgd(x: np.ndarray, beta: int) -> np.ndarray:
    # Sigmoid function
    return 1 / (np.exp(-beta * x) + 1)


def _swish(x: np.ndarray, beta: int) -> np.ndarray:
    # Swish function
    return x * _sgd(x, beta)



from sklvq import discriminants
from sklvq._utils import init_class
from sklvq.discriminants import DiscriminantBaseClass


def check_init_discriminant(discriminant_string):
    discriminant_class = init_class(discriminants, discriminant_string)

    assert isinstance(discriminant_class, type)

    activation_instance = discriminant_class()

    assert isinstance(activation_instance, DiscriminantBaseClass)

    return discriminant_class


# When there are discriminant functions with aliases re-enable
# def test_aliases():
#     for value in ALIASES.keys():
#         check_init_discriminant(value)



from .test_common import check_init_activation

import numpy as np


# Check if grab returns correct class, defaults of init (if any), basic workings are correct.
def test_identity():
    identity_class = check_init_activation("identity")

    identity = identity_class()

    rng = np.random.RandomState(0)
    x = rng.random_sample((5, 4))

    assert np.all(identity(x) == x)

    assert np.all(identity.gradient(x) == np.ones(x.shape))



from . import ScipyBaseSolver
from ..objectives import ObjectiveBaseClass


class BroydenFletcherGoldfarbShanno(ScipyBaseSolver):
    r"""Broyden Fletcher Goldfarb Shanno (BFGS)

    See the documentation of scipy for a complete parameter list and description.

    Parameters
    ----------
    jac: None
        Is set automatically to objective gradient method. However, if no gradient function
        is available, e.g., for a custom distance function, then jac can be set to None.
    callback: callable
        Differently from the non-scipy solvers the signature is callback(xk) with xk the current
        set of variables, which are the  model parameters flattened to one 1D array.
    """

    def __init__(self, objective: ObjectiveBaseClass, **kwargs):
        super(BroydenFletcherGoldfarbShanno, self).__init__(
            objective, method="BFGS", **kwargs
        )



import numpy as np

from scipy.spatial.distance import cdist

from ._base import DistanceBaseClass

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import GLVQ


class Euclidean(DistanceBaseClass):
    """Euclidean distance function

    Class that holds the euclidean distance function and its gradient.

    Parameters
    ----------
    force_all_finite  : {True, False, "allow-nan"}
        Parameter to indicate that NaNLVQ distance variant should be used. If true no nans are
        allowed. If False or "allow-nan" nans are allowed.

    See also
    --------
    SquaredEuclidean, AdaptiveSquaredEuclidean, LocalAdaptiveSquaredEuclidean

    Notes
    -----
    Compatible with the :class:`.GLVQ` algorithm (only).
    """

    __slots__ = ()

    def __call__(self, data: np.ndarray, model: "GLVQ") -> np.ndarray:
        r"""Computes the Euclidean distance:
            .. math::

                d(\mathbf{w}, \mathbf{x}) = \sqrt{(\mathbf{x} - \mathbf{w})^{\top} (\mathbf{x} - \mathbf{w})},

            with :math:`\mathbf{w}` a prototype and :math:`\mathbf{x}` a sample.

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distances to the prototypes of the model need to be
            computed.
        model : GLVQ
            A GLVQ model instance, containing the prototypes.

        Returns
        -------
        ndarray with shape (n_samples, n_prototypes)
            Evaluation of the distance between each sample and prototype of the model.
        """
        distance_function = "euclidean"
        if model.force_all_finite == "allow-nan":
            distance_function = _nan_euclidean

        return cdist(data, model.prototypes_, distance_function)

    def gradient(self, data: np.ndarray, model: "GLVQ", i_prototype: int) -> np.ndarray:
        r"""Computes the gradient of the euclidean distance with respect to a single
        prototype:

            .. math::
                \frac{\partial d}{\partial \mathbf{w}_i} = -2 \cdot (\mathbf{x} - \mathbf{w}_i)

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distance gradient to the prototypes of the model need to be
            computed.
        model : GLVQ
            A GLVQ model instance.
        i_prototype : int
            Index of the prototype to compute the gradient for.

        Returns
        -------
        ndarray with shape (n_samples, n_features)
            The gradient of the prototype with respect to every sample in the data.
        """
        prototype = model.get_model_params()[i_prototype, :]

        difference = data - prototype

        if model.force_all_finite == "allow-nan" or False:
            difference[np.isnan(difference)] = 0.0

        # Euclidean distance to single prototype. Equal to: np.sqrt(np.sum((data - prototype)**2))
        denominator = np.sqrt(np.einsum("ij, ij -> i", difference, difference))

        # Might happen if a sample is exactly on a prototype, mostly caused by nans in the data.
        denominator[denominator == 0.0] = 1.0

        distance_gradient = -1 * difference / denominator[:, np.newaxis]

        return distance_gradient


def _nan_euclidean(u: np.ndarray, v: np.ndarray) -> np.ndarray:
    # Euclidean distance between two vectors u and v, ignoring nans.
    return np.sqrt(np.nansum(u - v) ** 2)



from sklvq import activations
from sklvq._utils import init_class
from sklvq.activations import ALIASES
from sklvq.activations import ActivationBaseClass


def check_init_activation(activation_string):
    activation_class = init_class(activations, activation_string)

    assert isinstance(activation_class, type)

    activation_instance = activation_class()

    assert isinstance(activation_instance, ActivationBaseClass)

    return activation_class


def test_aliases():
    for value in ALIASES.keys():
        check_init_activation(value)



import numpy as np

from .test_common import check_init_activation

import pytest


def test_soft_plus():
    soft_plus_class = check_init_activation("soft-plus")

    assert isinstance(soft_plus_class, type)

    soft_plus = soft_plus_class(beta=2)

    assert soft_plus.beta == 2

    soft_plus = soft_plus_class()

    assert soft_plus.beta == 1

    wrong_betas = [-1, 0]
    for wrong_beta in wrong_betas:
        with pytest.raises(ValueError):
            soft_plus_class(beta=wrong_beta)

    # >0 at x = 0
    assert soft_plus(np.array([0])) > 0

    # >0 at x < 0
    assert np.all(soft_plus(np.array([-2, -4, -6, -8, -20, -50, -100]) > 0))

    # Increasing upto gradient of beta...
    assert np.max(soft_plus.gradient(np.array([0, 4, 8, 12, 20]))) == pytest.approx(
        soft_plus.beta
    )

    soft_plus = soft_plus_class(beta=10)

    # Increasing upto gradient of beta...
    assert np.max(soft_plus.gradient(np.array([0, 4, 8, 12, 20]))) == pytest.approx(
        soft_plus.beta
    )

    # soft plus gradient becomes nan when 1 + (beta * e^x) becomes inf. Which happens with larger x.
    # which should probably raise an error.

    # Potentially raising errors when division by 0.0 or inf happens....



import numpy as np
import pytest

from sklearn import datasets
from sklearn import preprocessing
from sklearn.model_selection import (
    GridSearchCV,
    RepeatedStratifiedKFold,
)
from sklearn.pipeline import make_pipeline

from .. import LGMLVQ


def test_lgmlvq_hyper_params():
    X, y = datasets.load_iris(return_X_y=True)

    model = LGMLVQ(prototype_n_per_class=6, relevance_localization="prototypes").fit(
        X, y
    )
    assert model.omega_.shape[0] == (model.classes_.size * 6)

    model = LGMLVQ(prototype_n_per_class=6, relevance_localization="class").fit(X, y)
    assert model.omega_.shape[0] == model.classes_.size

    with pytest.raises(ValueError):
        LGMLVQ(prototype_n_per_class=6, relevance_localization="abc").fit(X, y)

    with pytest.raises(ValueError):
        LGMLVQ(prototype_n_per_class=6, relevance_localization=8).fit(X, y)


def test_lgmlvq():
    iris = datasets.load_iris()

    estimator = LGMLVQ()
    pipeline = make_pipeline(preprocessing.StandardScaler(), estimator)

    # Run each solver ones
    solvers_types = [
        "lbfgs",
        "bfgs",
        "waypoint-gradient-descent",
        "adaptive-moment-estimation",
    ]
    stochastic_solver_types = [
        "steepest-gradient-descent",
    ]
    discriminant_types = ["relative-distance"]

    # Every compatible distance
    distance_types = ["local-adaptive-squared-euclidean"]

    # Every compatible activation
    activation_types = ["identity", "sigmoid", "soft-plus", "swish"]

    param_grid = [
        {
            "lgmlvq__solver_type": solvers_types,
            "lgmlvq__discriminant_type": discriminant_types,
            "lgmlvq__distance_type": distance_types,
            "lgmlvq__activation_type": activation_types,
        },
        {
            "lgmlvq__relevance_normalization": [True, False],
            "lgmlvq__solver_type": stochastic_solver_types,
            "lgmlvq__solver_params": [
                {"batch_size": 0, "step_size": np.array([0.1, 0.001])},
                {"batch_size": 1, "step_size": np.array([0.1, 0.01])},
                {"batch_size": 2, "step_size": np.array([0.1, 0.01])},
            ],
            "lgmlvq__discriminant_type": discriminant_types,
            "lgmlvq__distance_type": distance_types,
            "lgmlvq__activation_type": activation_types,
        },
    ]

    repeated_kfolds = RepeatedStratifiedKFold(n_splits=2, n_repeats=1)

    search = GridSearchCV(
        pipeline,
        param_grid,
        scoring="accuracy",
        cv=repeated_kfolds,
        return_train_score=True,
    )

    search.fit(iris.data, iris.target)

    assert np.all(search.cv_results_["mean_train_score"] > 0.75)
    assert np.all(search.cv_results_["mean_test_score"] > 0.75)

    print("\nBest parameter (CV score=%0.3f):" % search.best_score_)
    print(search.best_params_)



from . import ActivationBaseClass
import numpy as np

from typing import Union


class SoftPlus(ActivationBaseClass):
    """Soft+ function

    Class that holds the soft+ function and gradient as discussed in `[1]`_

    Parameters
    ----------
    beta : int or float, optional, default=1
           Positive non-zero value that controls the steepness of the Soft+ function.

    See also
    --------
    Identity, Sigmoid, Swish

    References
    ----------
    _`[1]` Villmann, T., Ravichandran, J., Villmann, A., Nebel, D., & Kaden, M. (2019). "Activation
    Functions for Generalized Learning Vector Quantization - A Performance Comparison", 2019.
    """

    __slots__ = "beta"

    def __init__(self, beta: Union[int, float] = 1):
        if beta <= 0:
            raise ValueError(
                "The activation function {} expects beta > 0, but got beta = {}".format(
                    type(self).__name__, beta
                )
            )

        self.beta = beta

    def __call__(self, x: np.ndarray) -> np.ndarray:
        r"""Implements the soft+ function:
            .. math::

                f(\mathbf{x}) = \ln(1 + e^{\beta \cdot \mathbf{x}})

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the soft+ function.
        """
        return np.log(1 + np.exp(self.beta * x))

    def gradient(self, x: np.ndarray) -> np.ndarray:
        r"""Implements the sigmoid function's gradient:
            .. math::

                \frac{\partial f}{\partial \mathbf{x}} = \frac{\beta \cdot e^{\beta \cdot \mathbf{x}}}{1 + e^{\beta \cdot \mathbf{x}}}

        Parameters
        ----------
        x : ndarray

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the soft+ function's gradient.
        """

        exp = np.exp(self.beta * x)
        return (self.beta * exp) / (1 + exp)



from importlib import import_module
from typing import List


def _parse_module_string(parameter_string: str) -> str:
    # Expects all lower case input and converts all hyphens "-" in underscores "_"
    return "_" + parameter_string.replace("-", "_")


def _parse_class_string(parameter_string: str) -> str:
    # Expect all lower case input and converts it to camel-case. The input is split on hyphens "-"
    return "".join(
        [substring.capitalize() for substring in parameter_string.rsplit("-")]
    )


def _parse_import_location_from_string(parameter_string: str) -> (str, str):
    # Converts everything to lower case
    parameter_string = parameter_string.casefold()

    # Return "correct" module and class strings
    return _parse_module_string(parameter_string), _parse_class_string(parameter_string)


def _import_class_from_string(package_string: str, class_string: str) -> type:
    module_string, class_string = _parse_import_location_from_string(class_string)

    try:
        module = import_module("." + module_string, package_string)
        specified_class = getattr(module, class_string)
    except (ModuleNotFoundError, AttributeError, TypeError):
        raise ImportError(
            "{}.{} is not part of our collection".format(module_string, class_string)
        )
    return specified_class


def _import_from_string(
    package,
    class_string: str,
    aliases: dict,
    param_name: str,
    valid_strings: List[str] = None,
) -> type:
    if class_string in aliases.keys():
        class_string = aliases[class_string]

    if valid_strings is not None:
        if not (class_string in valid_strings):
            raise ValueError(f"Provided {param_name} is invalid.")

    return _import_class_from_string(package, class_string)


def init_class(package, class_type, valid_class_types=None):
    if isinstance(class_type, str):
        parameter_class = package.import_from_string(class_type, valid_class_types)
    elif isinstance(class_type, type):
        parameter_class = class_type
    else:
        raise ValueError("Provided value '{}' is invalid.".format(class_type))
    return parameter_class



from . import DistanceBaseClass

import numpy as np
from scipy.spatial.distance import cdist

from ._adaptive_squared_euclidean import (
    _nan_mahalanobis,
    _prototype_gradient,
    _omega_gradient,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import LGMLVQ


class LocalAdaptiveSquaredEuclidean(DistanceBaseClass):
    """Local adaptive squared Euclidean distance

    Class that holds the localized adaptive squared Euclidean distance function and its gradient as
    described in `[1]`_ and `[2]`_.

    Parameters
    ----------
    force_all_finite  : {True, False, "allow-nan"}
        Parameter to indicate that NaNLVQ distance variant should be used. If true no nans are
        allowed. If False or "allow-nan" nans are allowed.

    See also
    --------
    Euclidean, SquaredEuclidean, AdaptiveSquaredEuclidean

    Notes
    -----
    Compatible with the :class:`.LGMLVQ` algorithm (only).

    References
    ----------
    _`[1]` Schneider, P. (2010). Advanced methods for prototype-based classification. Groningen.

    _`[2]` Schneider, P., Biehl, M., & Hammer, B. (2009). Adaptive Relevance Matrices in Learning
    Vector Quantization. Neural Computation, 21(12), 3532–3561.
    """

    __slots__ = ()

    def __call__(self, data: np.ndarray, model: "LGMLVQ") -> np.ndarray:
        r"""Computes the local variant of the adaptive squared Euclidean distance:

            .. math::
                d^{\Lambda}(\mathbf{w}, \mathbf{x}) = (\mathbf{x} - \mathbf{w})^{\top}
                \Omega_j^{\top} \Omega_j (\mathbf{x} - \mathbf{w})

        with :math:`\Omega_j` depending on the localization setting of the model and
        :math:`\Lambda_j = \Omega_j^{\top} \Omega_j`. The localization can be either per
        prototype or per class, see the documentation of :class:`.LGMLVQ`.

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The data for which the distance gradient to the prototypes of the model need to be
            computed.
        model : LGMLVQ
            A LGMLVQ model instance, containing the prototypes and relevance matrices.

        Returns
        -------
        ndarray with shape (n_samples, n_prototypes)
            Evaluation of the distance between each sample in the data and prototype of the model.
        """
        prototypes_, omegas_ = model.get_model_params()
        prototypes_labels_ = model.prototypes_labels_

        distance_function = "mahalanobis"
        kwarg_str = "VI"

        if model.force_all_finite == "allow-nan":
            distance_function = _nan_mahalanobis
            # RM because VI is filtered out of the  kwargs by cdist...
            kwarg_str = "RM"

        cdists = np.zeros((data.shape[0], model._prototypes_shape[0]))

        if model.relevance_localization == "prototypes":
            for i, (prototype, omega) in enumerate(zip(prototypes_, omegas_)):
                cdists[:, i] = cdist(
                    data,
                    np.atleast_2d(prototype),
                    distance_function,
                    **{kwarg_str: model._compute_lambda(omega)},
                ).squeeze()

        if model.relevance_localization == "class":
            for i, omega in enumerate(omegas_):
                # Prototype labels are indices to model.classes_ so all prototypes with 'index'
                # i as label have the same class.
                prototypes = prototypes_[i == prototypes_labels_, :]
                cdists[:, i == prototypes_labels_] = cdist(
                    data,
                    np.atleast_2d(prototypes),
                    distance_function,
                    **{kwarg_str: model._compute_lambda(omega)},
                )

        return cdists ** 2

    def gradient(
        self, data: np.ndarray, model: "LGMLVQ", i_prototype: int
    ) -> np.ndarray:
        r"""Computes the gradient of the localized adaptive squared euclidean distance function
        with respect to a specified prototype:

            .. math::
                \frac{\partial d}{\partial \mathbf{w}_i} = -2 \Lambda_j (\mathbf{x} - \mathbf{w}_i)

        and implicitly coupled omega matrix (per element):

            .. math::
                \frac{\partial d}{\partial \Omega_{lm}} =  2 \sum_i (x^i - w^i)
                \Omega_{li} (x^m - w^m)

        Parameters
        ----------
        data : ndarray with shape (n_samples, n_features)
            The X for which the distance gradient to the prototypes of the model need to be
            computed.
        model : LGMLVQ
            The LGMLVQ model instance, containing the prototypes and relevance matrices.
        i_prototype : int
            An integer index value of the relevant prototype

        Returns
        -------
        ndarray with shape (n_samples, n_features + n_omega_elements)
            The gradient of the prototype and omega matrix with respect to each data sample.
        """

        (prototypes, omegas) = model.get_model_params()

        prototype = prototypes[i_prototype, :]
        omega = omegas[model.prototypes_labels_[i_prototype], :, :]

        num_samples, _ = data.shape

        distance_gradient = np.empty(
            (num_samples, prototype.size + omega.size), dtype="float64", order="C"
        )

        # Difference we need for both the prototype and omega part.
        difference = data - prototype

        if model.force_all_finite == "allow-nan":
            difference[np.isnan(difference)] = 0.0

        _prototype_gradient(
            difference, omega, out=distance_gradient[:, : prototype.size]
        )

        distance_gradient_omega_view = distance_gradient[:, prototype.size :].reshape(
            (num_samples, *omega.shape)
        )

        # Omega gradient part
        _omega_gradient(difference, omega, out=distance_gradient_omega_view)

        return distance_gradient



from ._base import SolverBaseClass, ScipyBaseSolver
from ._steepest_gradient_descent import SteepestGradientDescent
from ._waypoint_gradient_descent import WaypointGradientDescent
from ._adaptive_moment_estimation import AdaptiveMomentEstimation
from ._limited_memory_bfgs import LimitedMemoryBfgs
from ._broyden_fletcher_goldfarb_shanno import BroydenFletcherGoldfarbShanno

from sklvq._utils import _import_from_string

__all__ = [
    "SolverBaseClass",
    "ScipyBaseSolver",
    "SteepestGradientDescent",
    "WaypointGradientDescent",
    "AdaptiveMomentEstimation",
    "LimitedMemoryBfgs",
    "BroydenFletcherGoldfarbShanno",
]

ALIASES = {
    "sgd": "steepest-gradient-descent",
    "bgd": "steepest-gradient-descent",
    "bfgs": "broyden-fletcher-goldfarb-shanno",
    "lbfgs": "limited-memory-bfgs",
    "adam": "adaptive-moment-estimation",
}


def import_from_string(class_string, valid_strings=None) -> type:
    return _import_from_string(
        __name__, class_string, ALIASES, "objective_type", valid_strings
    )



import numpy as np

from sklvq import distances
from sklvq._utils import init_class
from sklvq.distances import DistanceBaseClass
from sklvq.models import GMLVQ, LGMLVQ


def check_init_distance(distance_string):
    distatance_class = init_class(distances, distance_string)

    assert isinstance(distatance_class, type)

    distance_instance = distatance_class()

    assert isinstance(distance_instance, DistanceBaseClass)


# When there are distance functions with aliases re-enable
# def test_aliases():
#     for value in ALIASES.keys():
#         check_init_distance(value)


def check_distance(distfun, data, model):

    dists = distfun(data, model)
    # Samples by prototypes
    assert np.all(dists.shape == (data.shape[0], model.prototypes_.shape[0]))

    # 'Never' negative
    assert np.all(distfun(data, model) >= 0.0)

    # a to b is equal to b to a
    assert dists[1, 0] == dists[0, 1]

    # Same values in negative direction is same as the samples in the positive direction.
    assert dists[0, 0] == dists[1, 1]

    for i_prototype in range(0, model.prototypes_.shape[1]):
        # Check if shape of the gradient makes sense
        gradient = distfun.gradient(data, model, 0)

        gradient_size = model.prototypes_[0, :].size

        if isinstance(model, (GMLVQ, LGMLVQ)):
            if model.omega_.ndim == 3:
                gradient_size += model.omega_[0, :, :].size
            else:
                gradient_size += model.omega_.size

        # Size of num_samples x single prototype...
        assert gradient.shape == (data.shape[0], gradient_size)

        # Happens when X is exactly on the prototype for some distances.
        assert np.all(np.logical_not(np.isnan(gradient)))

        # This should also not happen
        assert np.all(np.logical_not(np.isinf(gradient)))



import numpy as np

from .test_common import check_distance
from .test_common import check_init_distance

from sklvq.models import LGMLVQ


def test_adaptive_squared_euclidean():
    check_init_distance("local-adaptive-squared-euclidean")

    # Some X and prototypes
    data = np.array([[1, 2, 3], [-1, -2, -3], [0, 0, 0]], dtype="float64")
    p = np.array([[1, 2, 3], [-1, -2, -3], [0, 0, 0]])

    shape = (data.shape[1], data.shape[1])
    o = np.array([np.eye(*shape) for _ in range(p.shape[0])])

    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean",
        relevance_localization="prototypes",
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(o)

    check_distance(model._distance, data, model)

    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean", relevance_localization="class"
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(o)

    check_distance(model._distance, data, model)

    # Rectangular omega
    orec = np.array([np.eye(*shape)[:2, :] for _ in range(p.shape[0])])
    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean",
        relevance_localization="prototypes",
        relevance_n_components=2,
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(orec)

    check_distance(model._distance, data, model)

    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean",
        relevance_localization="class",
        relevance_n_components=2,
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(orec)

    check_distance(model._distance, data, model)

    # Check force_all_finite settings
    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean",
        relevance_localization="prototypes",
        force_all_finite="allow-nan",
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(o)

    nan_data = np.copy(data)
    nan_data[0, 0] = np.nan
    nan_data[1, 0] = np.nan

    check_distance(model._distance, nan_data, model)

    model = LGMLVQ(
        distance_type="local-adaptive-squared-euclidean",
        relevance_localization="class",
        force_all_finite="allow-nan",
    )
    model.fit(data, np.array([0, 1, 2]))
    model.set_prototypes(p)
    model.set_omega(o)

    check_distance(model._distance, nan_data, model)



# Typing
from typing import Union

import numpy as np

from . import LVQBaseClass
from ..objectives import GeneralizedLearningObjective

ModelParamsType = np.ndarray

DISTANCES = [
    "euclidean",
    "squared-euclidean",
]

SOLVERS = [
    "adaptive-moment-estimation",
    "broyden-fletcher-goldfarb-shanno",
    "limited-memory-bfgs",
    "steepest-gradient-descent",
    "waypoint-gradient-descent",
]


class GLVQ(LVQBaseClass):
    r"""Generalized Learning Vector Quantization

    This model uses the :class:`sklvq.objectives.GeneralizedLearningObjective` as its objective
    function `[1]`_.

    Parameters
    ----------
    distance_type : {"squared-euclidean", "euclidean"} or Class, default="squared-euclidean"
        The distance function. Can be one from the following list or a custom class:

        - "squared-euclidean"
            See :class:`sklvq.distances.SquaredEuclidean`
        - "euclidean"
            See :class:`sklvq.distances.Euclidean`

    distance_params : dict, optional, default=None
        Parameters passed to init of distance class.

    activation_type : {"identity", "sigmoid", "soft+", "swish"} or Class, default="sigmoid"
        The activation function used in the objective function. Can be any of the activation
        function in the list or custom class.

        - "identity"
            See :class:`sklvq.activations.Identity`
        - "sigmoid"
            See :class:`sklvq.activations.Sigmoid`
        - "soft+"
            See :class:`sklvq.activations.SoftPlus`
        - "swish"
            See :class:`sklvq.activations.Swish`

    activation_params : dict, default=None
        Parameters passed to init of activation function. See the documentation of the activation
        functions for parameters and defaults.

    discriminant_type : "relative-distance" or Class
        The discriminant function.  Note that different discriminant type may require to rewrite
        the ``decision_function`` and ``predict_proba`` methods.

        - "relative-distance"
            See :class:`sklvq.discriminants.RelativeDistance`

    discriminant_params : dict, default=None
        Parameters passed to init of discriminant callable. See the documentation of the
        discriminant functions for parameters and defaults.

    solver_type : {"sgd", "wgd", "adam", "lbfgs", "bfgs"},
        The solver used for optimization

        - "sgd" or "steepest-gradient-descent"
            See :class:`sklvq.solvers.SteepestGradientDescent`.
        - "wgd" or "waypoint-gradient-descent"
            See :class:`sklvq.solvers.WaypointGradientDescent`.
        - "adam" or "adaptive-moment-estimation"
            See :class:`sklvq.solvers.AdaptiveMomentEstimation`.
        - "bfgs" or "broyden-fletcher-goldfarb-shanno"
            See :class:`sklvq.solvers.BroydenFletcherGoldfarbShanno`
        - "lbfgs" or "limited-memory-bfgs"
            See :class:`skvlq.solvers.LimitedMemoryBfgs`

    solver_params : dict, default=None
        Parameters passed to init of solvers. See the documentation of the solvers relevant
        parameters and defaults.

    prototype_init: "class-conditional-mean" or ndarray, default="class-conditional-mean"
        Default will initiate the prototypes to the class conditional mean with a small random
        offset. Custom numpy array can be passed to change the initial positions of the prototypes.

    prototype_n_per_class: int or np.ndarray, optional, default=1
        Default will generate single prototype per class. In the case of unequal number of
        prototypes per class is needed, provide this as np.ndarray. For example,
        prototype_n_per_class = np.array([1, 6, 3]) this will result in one prototype for the first class,
        six for the second, and three for the third. Note that the order needs to be the same as the on in the
        classes\_ attribute, which is equal to calling np.unique(labels).

    random_state : int, RandomState instance, default=None
        Set the random number generation for reproducibility purposes. Used in random offset of prototypes and
        shuffling of the data in the solvers.

    force_all_finite : {True, "allow-nan"}, default=True
        Whether to raise an error on np.inf, np.nan, pd.NA in array. The possibilities are:

        - True: Force all values of array to be finite.
        - "allow-nan": accepts only np.nan and pd.NA values in array. Values cannot be infinite.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        The original and unique labels found in the data.

    prototypes_ : ndarray of shape (n_protoypes, n_features)
        Positions of the prototypes after fit(X, labels) has been called.

    prototypes_labels_ : ndarray of shape (n_prototypes)
        Labels for each prototypes. Labels are indexes to ``classes_``

    References
    ----------
    _`[1]` Sato, A., and Yamada, K. (1996) "Generalized Learning Vector Quantization."
    Advances in Neural Network Information Processing Systems, 423–429, 1996.
    """

    classes_: np.ndarray
    prototypes_: np.ndarray
    prototypes_labels_: np.ndarray

    def __init__(
        self,
        distance_type: Union[str, type] = "squared-euclidean",
        distance_params: dict = None,
        activation_type: Union[str, type] = "sigmoid",
        activation_params: dict = None,
        discriminant_type: Union[str, type] = "relative-distance",
        discriminant_params: dict = None,
        solver_type: Union[str, type] = "steepest-gradient-descent",
        solver_params: dict = None,
        prototype_init: str = "class-conditional-mean",
        prototype_n_per_class: Union[int, np.ndarray] = 1,
        random_state: Union[int, np.random.RandomState] = None,
        force_all_finite: Union[str, bool] = True,
    ):
        self.activation_type = activation_type
        self.activation_params = activation_params
        self.discriminant_type = discriminant_type
        self.discriminant_params = discriminant_params

        super(GLVQ, self).__init__(
            distance_type,
            distance_params,
            DISTANCES,
            solver_type,
            solver_params,
            SOLVERS,
            prototype_init,
            prototype_n_per_class,
            random_state,
            force_all_finite,
        )

    ###########################################################################################
    # The "Getter" and "Setter" that are used by the solvers to set and get model params.
    ###########################################################################################

    def get_model_params(self) -> ModelParamsType:
        """
        Returns a view of all model parameters, which are only the prototypes.

        Returns
        -------
        ndarray
             Returns a view of the prototypes as ndarray.
        """
        return self.get_prototypes()

    def set_model_params(self, new_model_params: ModelParamsType) -> None:
        """
        Changes the model's internal parameters. Copies the values of model_params into
        ``self.prototypes_`` therefor updating the ``self.variables_`` array.

        Parameters
        ----------
        new_model_params : ndarray of shape (n_prototypes, n_features)
            In the   case the prototypes.
        """
        self.set_prototypes(new_model_params)

    ###########################################################################################
    # Functions to transform the 1D variables array to model parameters and back
    ###########################################################################################

    def to_model_params_view(self, var_buffer: np.ndarray) -> ModelParamsType:
        """
        Should create a view of the variables array in prototype shape.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray
            Returns the prototypes as ndarray.
        """
        return self.to_prototypes_view(var_buffer)

    def to_prototypes_view(self, var_buffer: np.ndarray) -> np.ndarray:
        """
        Returns the prototypes into the provided var_buffer. I.e., it selects/views the
        appropriate part of memory and reshapes it.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray of shape (n_prototypes, n_features)
            Prototype view into the var_buffer.
        """
        return var_buffer.reshape(self._prototypes_shape)

    ###########################################################################################
    # Solver Normalization functions
    ###########################################################################################

    def normalize_variables(self, var_buffer: np.ndarray) -> None:
        r"""
        Modifies the var_buffer as if it was the variables array provided
        by ``get_variables()``. As variables only contain prototypes it will now  contain the
        normalized prototypes.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray
            Same shape and size as input, but normalized.
        """
        LVQBaseClass._normalize_prototypes(self.to_prototypes_view(var_buffer))

    ###########################################################################################
    # Solver helper functions
    ###########################################################################################

    def add_partial_gradient(self, gradient, partial_gradient, i_prototype) -> None:
        """
        Adds the partial gradient to the correct part of the gradient, which  depends on
        ``i_prototype``.

        Parameters
        ----------
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.

        partial_gradient : ndarray
            1d array containing the partial gradient.

        i_prototype : int
            The index of the prototype to which the partial gradient was  computed.
        """
        n_features = self.n_features_in_

        prots_view = self.to_prototypes_view(gradient)
        np.add(
            prots_view[i_prototype, :],
            partial_gradient[:n_features],
            out=prots_view[i_prototype, :],
        )

    def mul_step_size(self, step_size: Union[int, float], gradient: np.ndarray) -> None:
        """
        As GLVQ only has prototypes that are optimized the step_size should be a single float
        and can just be used to multiply the gradient inplace.

        Parameters
        ----------
        step_size : float or ndarray
            The scalar or list of values containing the step sizes.
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.
        """
        gradient *= step_size

    ###########################################################################################
    # Initialization required functions
    ###########################################################################################

    def _init_variables(self) -> None:
        self._variables = np.empty(self._prototypes_size, dtype="float64", order="C")

    def _check_model_params(self):
        self._check_prototype_params()

    def _init_model_params(self, X, y) -> None:
        self._init_prototypes(X, y)

    def _init_objective(self):
        self._objective = GeneralizedLearningObjective(
            self.activation_type,
            self.activation_params,
            self.discriminant_type,
            self.discriminant_params,
        )



from typing import Tuple, Union, List

import numpy as np
from sklearn.utils.validation import check_is_fitted, check_array

from . import LVQBaseClass
from ..objectives import GeneralizedLearningObjective

ModelParamsType = Tuple[np.ndarray, np.ndarray]

DISTANCES = [
    "local-adaptive-squared-euclidean",
]

SOLVERS = [
    "adaptive-moment-estimation",
    "broyden-fletcher-goldfarb-shanno",
    "limited-memory-bfgs",
    "steepest-gradient-descent",
    "waypoint-gradient-descent",
]


class LGMLVQ(LVQBaseClass):
    r"""Localized Generalized Matrix Learning Vector Quantization

    This model uses the :class:`sklvq.objectives.GeneralizedLearningObjective` as its objective
    function `[1]`_. In addition to learning the positions of the prototypes it learns a set
    of relevance matrices in a localized manner, that are used in the distance functions `[2]`_.

    Parameters
    ----------
    distance_type : "local-adaptive-squared-euclidean" or Class
        Distance function that employs multiple relevance matrix in its calculation. This is
        controlled by the localization setting.

        - "local-adaptive-squared-euclidean"
            See :class:`sklvq.distances.LocalAdaptiveSquaredEuclidean`

    distance_params : dict, default=None
        Parameters passed to init of distance class.

    activation_type : {"identity", "sigmoid", "soft+", "swish"} or Class, default="sigmoid"
        The activation function used in the objective function. Can be any of the activation
        function in the list or custom class.

        - "identity"
            See :class:`sklvq.activations.Identity`
        - "sigmoid"
            See :class:`sklvq.activations.Sigmoid`
        - "soft+"
            See :class:`sklvq.activations.SoftPlus`
        - "swish"
            See :class:`sklvq.activations.Swish`


    activation_params : Dict, default=None
        Parameters passed to init of activation function. See the documentation of the activation
        functions for parameters and defaults.

    discriminant_type : "relative-distance" or Class
        The discriminant function.  Note that different discriminant type may require to rewrite
        the ``decision_function`` and ``predict_proba`` methods.

        - "relative-distance"
            See :class:`sklvq.discriminants.RelativeDistance`

    discriminant_params : Dict, default=None
        Parameters passed to init of discriminant callable. See the documentation of the
        discriminant functions for parameters and defaults.

    solver_type : {"sgd", "wgd", "adam", "lbfgs", "bfgs"},
        The solver used for optimization

        - "sgd" or "steepest-gradient-descent"
            See :class:`sklvq.solvers.SteepestGradientDescent`.
        - "wgd" or "waypoint-gradient-descent"
            See :class:`sklvq.solvers.WaypointGradientDescent`.
        - "adam" or "adaptive-moment-estimation"
            See :class:`sklvq.solvers.AdaptiveMomentEstimation`.
        - "bfgs" or "broyden-fletcher-goldfarb-shanno"
            Implementation from scipy package.
        - "lbfgs" or "limited-memory-bfgs"
            Implementation from scipy package.

    solver_params : dict, default=None
        Parameters passed to init of solvers. See the documentation of the solvers relevant
        parameters and defaults.

    prototype_init: "class-conditional-mean" or ndarray, default="class-conditional-mean"
        Default will initiate the prototypes to the class conditional mean with a small random
        offset. Custom numpy array can be passed to change the initial positions of the prototypes.

    prototype_n_per_class: int or np.ndarray, optional, default=1
        Default will generate single prototype per class. In the case of unequal number of
        prototypes per class is needed, provide this as np.ndarray. For example,
        prototype_n_per_class = np.array([1, 6, 3]) this will result in one prototype for the first class,
        six for the second, and three for the third. Note that the order needs to be the same as the on in the
        classes\_ attribute, which is equal to calling np.unique(labels).

    relevance_init : {"identity", "random"} or np.ndarray, default="identity"
       Default will initiate the omega matrices to be the identity matrix. The rank of the matrix can be reduced by
        setting the ``relevance_n_components`` attribute `[3]`_.

    relevance_normalization: bool, optional, default=True
        Flag to indicate whether to normalize omega, whenever it is updated, such that the trace of the relevance matrix
        is equal to 1.

    relevance_n_components: str {"all"} or int, optional, default="all"
        For a square relevance matrix use the string "all" (default). For a rectangular relevance matrix use set the
        number of components explicitly by providing it as an int.

    relevance_localization: {"prototypes", "class"}, default="prototypes"
        Setting that controls the localization of the relevance matrices. Either per prototype,
        where each prototype has its own relevance matrix. Or per class where each class has its
        own relevance matrix. Note that when one prototype per class is used, changing this setting has no effect.

    random_state : int, RandomState instance, default=None
        Set the random number generation for reproducibility purposes. Used in random offset of prototypes and
        shuffling of the data in the solvers. Potentially, also used in the random generation of relevance matrix.

    force_all_finite : {True, "allow-nan"}, default=True
        Whether to raise an error on np.inf, np.nan, pd.NA in array. The possibilities are:

        - True: Force all values of array to be finite.
        - "allow-nan": accepts only np.nan and pd.NA values in array. Values cannot be infinite.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        Class labels for each output.

    prototypes_ : ndarray of shape (n_protoypes, n_features)
        Positions of the prototypes after ``fit(X, labels)`` has been called.

    prototypes_labels_ : ndarray of shape (n_prototypes)
        Labels for each prototypes. Labels are indexes to ``classes_``

    omega_: ndarray with size (n_matrices, n_features, n_features)
        ``omega_`` matrices that were found during training and define the relevance matrices
        ``lambda_``.

    lambda_: ndarray of size (n_matrices, n_features, n_features)
        The relevance matrices ``omega_.T.dot(omega_)`` per matrix.

    omega_hat_: ndarray
        The omega matrices found by the eigenvalue decomposition of the relevance matrices
        ``lambda_``. The eigenvectors (columns of ``omega_hat_``) can be used to transform the data
         `[3]`_. This results in multiple possible transformations, one per relevance matrix.

    eigenvalues_: ndarray
        The corresponding eigenvalues to ``omega_hat_`` found by the eigenvalue decomposition of
        the relevance matrices ``lambda_``

    References
    ----------
    _`[1]` Sato, A., and Yamada, K. (1996) "Generalized Learning Vector Quantization."
    Advances in Neural Network Information Processing Systems, 423–429, 1996.

    _`[2]` Schneider, P., Biehl, M., & Hammer, B. (2009). "Adaptive Relevance Matrices in
    Learning Vector Quantization" Neural Computation, 21(12), 3532–3561, 2009.

    _`[3]` Bunte, K., Schneider, P., Hammer, B., Schleif, F.-M., Villmann, T., & Biehl, M. (2012).
    "Limited Rank Matrix Learning, discriminative dimension reduction and visualization." Neural
    Networks, 26, 159–173, 2012.
    """

    classes_: np.ndarray
    prototypes_: np.ndarray
    prototypes_labels_: np.ndarray
    omega_: np.ndarray
    lambda_: np.ndarray
    omega_hat_: np.ndarray
    eigenvalues_: np.ndarray

    def __init__(
        self,
        distance_type: Union[str, type] = "local-adaptive-squared-euclidean",
        distance_params: dict = None,
        activation_type: Union[str, type] = "identity",
        activation_params: dict = None,
        discriminant_type: Union[str, type] = "relative-distance",
        discriminant_params: dict = None,
        solver_type: Union[str, type] = "steepest-gradient-descent",
        solver_params: dict = None,
        prototype_init: Union[str, np.ndarray] = "class-conditional-mean",
        prototype_n_per_class: [int, np.ndarray] = 1,
        relevance_init: Union[str, np.ndarray] = "identity",
        relevance_normalization: bool = True,
        relevance_n_components: Union[str, int] = "all",
        relevance_localization: str = "prototypes",
        random_state: Union[int, np.random.RandomState] = None,
        force_all_finite: Union[str, int] = True,
    ):
        self.activation_type = activation_type
        self.activation_params = activation_params
        self.discriminant_type = discriminant_type
        self.discriminant_params = discriminant_params
        self.relevance_init = relevance_init
        self.relevance_normalization = relevance_normalization
        self.relevance_n_components = relevance_n_components
        self.relevance_localization = relevance_localization

        super(LGMLVQ, self).__init__(
            distance_type,
            distance_params,
            DISTANCES,
            solver_type,
            solver_params,
            SOLVERS,
            prototype_init,
            prototype_n_per_class,
            random_state,
            force_all_finite,
        )

    ###########################################################################################
    # The "Getter" and "Setter" that are used by the solvers to set and get model params.
    ###########################################################################################

    def set_variables(self, new_variables: np.ndarray) -> None:
        """
        Modifies the ``self._variables`` by copying the values of ``new_variables`` into the
        memory of ``self._variables``.

        Parameters
        ----------
        new_variables : ndarray
            1d numpy array that contains all the model parameters in continuous memory
        """
        np.copyto(self._variables, new_variables)
        if self.relevance_normalization:
            LGMLVQ._normalize_omega(self.to_omega(self._variables))

    def set_model_params(self, new_model_params: ModelParamsType):
        """
        Changes the model's internal parameters. Copies the values of model_params into
        ``self.prototypes_`` and ``self.omega_`` therefore updating the ``self.variables_``
        array.

        Parameters
        ----------
        new_model_params : tuple of ndarrays
            Shapes depend on  initialization but in the case of a square relevance matrix:
            tuple((n_prototypes, n_features), (n_matrices, n_features, n_features))
        """
        new_prototypes, new_omega = new_model_params

        self.set_prototypes(new_prototypes)
        self.set_omega(new_omega)

        if self.relevance_normalization:
            LGMLVQ._normalize_omega(self.omega_)

    def get_model_params(self) -> ModelParamsType:
        """
        Returns a tuple of all model parameters. In this case the prototypes and omega matrix.

        Returns
        -------
        ndarray
             Returns a tuple of views, i.e., the prototypes and omega matrix.
        """
        return self.prototypes_, self.omega_

    ###########################################################################################
    # Specific "getters" and "setters" for the prototypes shared by every LVQ model.
    ###########################################################################################

    def get_omega(self):
        """
        Function to return ``self.omega_`` (consistency)

        Returns
        -------
            ndarray, with shape depending on initialization of omega.
        """
        return self.omega_

    def set_omega(self, omega):
        """
        Convenience function that makes sure to copy the value to ``self.omega_`` and not overwrite
        it.

        Parameters
        ----------
        omega : ndarray with same shape as ``self.omega_``
        """
        np.copyto(self.omega_, omega)

    ###########################################################################################
    # Functions to transform the 1D variables array to model parameters and back
    ###########################################################################################

    def to_model_params_view(self, var_buffer: np.ndarray) -> ModelParamsType:
        """
        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        tuple
            Returns a tuple with the prototypes and omega matrices as ndarrays.
        """
        return (
            self.to_prototypes_view(var_buffer),
            self.to_omega(var_buffer),
        )

    def to_prototypes_view(self, var_buffer: np.ndarray) -> np.ndarray:
        """
        Returns a view (of the shape of the model's prototypes) into the provided variables
        buffer of the same size as the model's variables array.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray of shape (n_prototypes, n_features)
            Prototype view into the var_buffer.
        """
        return var_buffer[: self._prototypes_size].reshape(self._prototypes_shape)

    def to_omega(self, var_buffer: np.ndarray) -> np.ndarray:
        """
        Returns a view (of the shape of the model's omega) into the provided variables
        buffer of the same size as the model's variables array.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.

        Returns
        -------
        ndarray
            Shape depending on initialization but in case of a square matrix (n_matrices,
            n_features, n_features).
        """
        return var_buffer[self._prototypes_size :].reshape(self._relevances_shape)

    ###########################################################################################
    # Solver Normalization functions
    ###########################################################################################

    def normalize_variables(self, var_buffer: np.ndarray) -> None:
        """
        Modifies the var_buffer as if it was the variables array provided
        by ``get_variables()``. Will select, reshape and normalize the correct parts of the
        variable buffer.

        Parameters
        ----------
        var_buffer : ndarray
            Array with the same size as the model's variables array as returned
            by ``get_variables()``.
        """
        (prototypes, omega) = self.to_model_params_view(var_buffer)

        self._normalize_prototypes(prototypes)
        self._normalize_omega(omega)

    @staticmethod
    def _normalize_omega(omega: np.ndarray) -> None:
        np.divide(
            omega,
            np.sqrt(np.einsum("ikj, ikj -> i", omega, omega)).reshape(
                omega.shape[0], 1, 1
            ),
            out=omega,
        )
        # denominator = np.sqrt(np.einsum("ikj, ikj -> i", omega, omega)).reshape(
        #     omega.shape[0], 1, 1
        # )
        # omega /= denominator

    ###########################################################################################
    # Solver helper functions
    ###########################################################################################

    def add_partial_gradient(self, gradient, partial_gradient, i_prototype) -> None:
        """
        Adds the partial gradient to the correct part of the gradient, which  depends on
        ``i_prototype``.

        Parameters
        ----------
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.

        partial_gradient : ndarray
            1d array containing the partial gradient.

        i_prototype : int
            The index of the prototype to which the partial gradient was  computed.
        """
        n_features = self.n_features_in_

        prots_view = self.to_prototypes_view(gradient)
        np.add(
            prots_view[i_prototype, :],
            partial_gradient[:n_features],
            out=prots_view[i_prototype, :],
        )
        omega_view = self.to_omega(gradient)[
            self.prototypes_labels_[i_prototype], :, :
        ].ravel()

        np.add(omega_view, partial_gradient[n_features:], out=omega_view)

    def mul_step_size(
        self, step_sizes: Union[int, float, np.ndarray], gradient: np.ndarray
    ) -> None:
        """
        If step sizes is a scalar value just multiplies the gradient with the step size. If it
        is an array (with same length as number of model parameters) each model parameter is
        multiplied by its own step size.

        Parameters
        ----------
        step_sizes : float or ndarray
            The scalar or list of values containing the step sizes.
        gradient : ndarray
            Same shape as the ``get_variables()`` would return.
        """
        if isinstance(step_sizes, int) | isinstance(step_sizes, float):
            gradient *= step_sizes
            return

        if isinstance(step_sizes, np.ndarray):
            if step_sizes.size > 1:
                prototypes, omegas = self.to_model_params_view(gradient)
                prototypes *= step_sizes[0]
                omegas *= step_sizes[1]

    ###########################################################################################
    # Initialization function
    ###########################################################################################

    def _init_variables(self) -> None:
        self._variables = np.empty(
            self._prototypes_size + self._relevances_size,
            dtype="float64",
            order="C",
        )

    def _check_model_params(self):
        self._check_prototype_params()
        self._check_relevances_params()

    def _init_model_params(self, X, y) -> None:
        self._init_prototypes(X, y)
        self._init_relevances()

    def _check_relevances_params(self):
        relevance_normalization = self.relevance_normalization
        if not isinstance(relevance_normalization, bool):
            raise ValueError("Provided normalization is invalid.")

        relevance_n_components = self.relevance_n_components
        if isinstance(relevance_n_components, str):
            if relevance_n_components == "all":
                shape = (self.n_features_in_, self.n_features_in_)
            else:
                raise ValueError("Provided n_components is invalid.")
        elif isinstance(relevance_n_components, int):
            if (
                self.relevance_n_components >= 1
                and relevance_n_components <= self.n_features_in_
            ):
                shape = (relevance_n_components, self.n_features_in_)
            else:
                raise ValueError("Provided n_components is invalid.")
        else:
            raise ValueError("Provided n_components is invalid.")

        relevance_localization = self.relevance_localization
        if isinstance(relevance_localization, str):
            if relevance_localization == "prototypes":
                self._relevances_shape = (self._prototypes_shape[0], *shape)
            elif relevance_localization == "class":
                self._relevances_shape = (self.classes_.size, *shape)
            else:
                raise ValueError("Provided localization is invalid.")
        else:
            raise ValueError("Provided localization is invalid.")

        self._relevances_size = np.prod(self._relevances_shape)

    def _init_relevances(self):
        self.omega_ = self.to_omega(self._variables)

        if isinstance(self.relevance_init, str):
            if self.relevance_init == "identity":
                identity = np.eye(*self._relevances_shape[1:])
                self.set_omega(
                    np.repeat(
                        identity[np.newaxis, :, :], self._relevances_shape[0], axis=0
                    )
                )
            elif self.relevance_init == "random":
                self.set_omega(
                    self.random_state_.uniform(
                        low=0, high=1, size=self._relevances_shape
                    )
                )
            else:
                raise ValueError("Provided relevance_init is invalid.")
        else:
            raise ValueError("Provided relevance_init is invalid.")

        if self.relevance_normalization:
            LGMLVQ._normalize_omega(self.omega_)

    def _init_objective(self):
        self._objective = GeneralizedLearningObjective(
            self.activation_type,
            self.activation_params,
            self.discriminant_type,
            self.discriminant_params,
        )

    ###########################################################################################
    # Other Algorithm specific stuff.
    ###########################################################################################

    def _after_fit(self, X: np.ndarray, y: np.ndarray):
        self.lambda_ = LGMLVQ._compute_lambdas(self.omega_)

        # Eigenvalues and column eigenvectors returned in ascending order
        eigenvalues, omega_hat = np.linalg.eigh(self.lambda_)

        self.eigenvalues_ = np.flip(eigenvalues, axis=1)
        self.omega_hat_ = np.transpose(np.flip(omega_hat, axis=2), axes=(0, 2, 1))

    @staticmethod
    def _compute_lambda(omega):
        # Equivalent to omega.T.dot(omega)
        return np.einsum("ji, jk -> ik", omega, omega)

    @staticmethod
    def _compute_lambdas(omegas):
        # Equivalent to omega.T.dot(omega) per omega (quite slow)
        return np.einsum("ikj, ikl -> ijl", omegas, omegas)

    ###########################################################################################
    # Transformer related functions
    ###########################################################################################

    def fit_transform(self, X: np.ndarray, y: np.ndarray, **trans_params) -> np.ndarray:
        r"""
        Parameters
        ----------
        X : ndarray with shape (n_samples, n_features)
            Data used for fit and that will be transformed.
        y : np.ndarray with length (n_samples)
            Labels corresponding to the X samples.
        trans_params :
            Parameters passed to transform function

        Returns
        -------
        The data projected on columns of ``omega_hat_`` with shape (n_matrices, n_samples,
        n_columns)
        """
        return self.fit(X, y).transform(X, **trans_params)

    def transform(
        self,
        X: np.ndarray,
        scale: bool = False,
        omega_hat_index: Union[int, List[int]] = 0,
    ) -> np.ndarray:
        r"""
        Parameters
        ----------
        X : np.ndarray with shape (n_samples, n_features)
            Data that needs to be transformed
        scale : {True, False}, default = False
            Controls if the eigenvectors the data is projected on are scaled by the square root
            of their eigenvalues.
        omega_hat_index : int or list
            The indices of the omega\_hats\_ the transformation should be computed for.

        Returns
        -------
        The data projected on columns of ``omega_hat_`` with shape (n_samples,
        n_columns, n_matrices)
        """
        check_is_fitted(self)

        X = check_array(X)

        transformation_matrix = self.omega_hat_[omega_hat_index, :, :]
        if transformation_matrix.ndim != 3:
            transformation_matrix = transformation_matrix[None, :, :]

        if scale:
            eigenvalues = np.sqrt(np.absolute(self.eigenvalues_[omega_hat_index, :]))

            transformation_matrix = (
                np.expand_dims(eigenvalues, axis=eigenvalues.ndim)
                * transformation_matrix
            )

        transformed_data = np.einsum("in, jmn -> imj", X, transformation_matrix)

        return transformed_data.squeeze()

    def _more_tags(self):
        # For some reason lgmlvq (with default settings) does not perform well on one of the test
        # cases build into sklearn's test cases. Which is something to look into, but this "fixes"
        # it in the mean time...
        return {"poor_score": True}



import numpy as np

from .test_common import check_init_discriminant


def test_relative_distance():
    rel_dist_class = check_init_discriminant("relative-distance")

    rel_dist = rel_dist_class()

    # Needs to be negative when dist_diff is bigger. Positive when dist_same is bigger and 0 if
    # they are the same:
    rng = np.random.RandomState(0)
    x = rng.random_sample((100, 1))
    y = rng.random_sample((100, 1))

    # Positive when x > y
    assert np.all((x > y) == (rel_dist(x, y) > 0))

    # Negative when y > x
    assert np.all((y > x) == (rel_dist(x, y) < 0))

    # All the same
    assert np.all((rel_dist(x, x) == 0.0) & (rel_dist(y, y) == 0.0))

    # Potentially raising errors when division by 0.0 or inf happens....

    # Potentially write gradient test cases...
    #   Check if same reverse of diff? Not sure...



import numpy as np

from ..objectives._base import ObjectiveBaseClass
from .. import activations, discriminants

from typing import Union
from typing import TYPE_CHECKING

from .._utils import init_class

if TYPE_CHECKING:
    from ..models import LVQBaseClass


ACTIVATION_FUNCTIONS = [
    "identity",
    "sigmoid",
    "soft-plus",
    "swish",
]

DISCRIMINANT_FUNCTIONS = [
    "relative-distance",
]


class GeneralizedLearningObjective(ObjectiveBaseClass):
    """Generalized learning objective

    Class that holds the generalized learning objective function and its gradient as described
    in `[1]`_.

    Parameters
    ----------
    activation_type : {"identity", "sigmoid", "soft-plus", "swish"} or type
        If string needs to be one of the indicated options. If not a string needs to be a custom
        activation class. See :class:`sklvq.activations.ActivationBaseClass`.
    activation_params : dict or None
        The dictionary with the parameters for the activation function or None if it doesn't
        require any parameters.
    discriminant_type: {"relative-distance"} or type
        Can only be the relative distance. If not a string it can be a custom class.
        See :class:`sklvq.discriminants.DiscriminantBaseClass`.
    discriminant_params : dict or None
        The dictionary with the parameters for the discriminant function or None if it doesn't
        require any parameters.

    Notes
    -----
    Compatible and used within the following models: :class:`.GLVQ`, :class:`.GMLVQ`,
    and :class:`.LGMLVQ`.

    References
    ----------
    _`[1]` Sato, A., and Yamada, K. (1996) "Generalized Learning Vector Quantization."
    Advances in Neural Network Information Processing Systems, 423–429, 1996."""

    def __init__(
        self,
        activation_type: Union[str, type],
        activation_params: dict,
        discriminant_type: Union[str, type],
        discriminant_params: dict,
    ):
        if activation_params is None:
            activation_params = {}

        activation_class = init_class(
            activations, activation_type, ACTIVATION_FUNCTIONS
        )
        self.activation = activation_class(**activation_params)

        if discriminant_params is None:
            discriminant_params = {}

        discriminant_class = init_class(
            discriminants, discriminant_type, DISCRIMINANT_FUNCTIONS
        )
        self.discriminant = discriminant_class(**discriminant_params)

    def __call__(
        self,
        model: "LVQBaseClass",
        data: np.ndarray,
        labels: np.ndarray,
    ) -> np.ndarray:
        r"""Computes the generalized learning objective:

            .. math::

                E = \sum_{i=1}^{N} f(\mu(d_0(\mathbf{x}_i), d_1(\mathbf{x}_i))

        with :math:`\mu(\cdot)` the discriminative function, :math:`f(\cdot)` the activation
        function, and :math:`d_0(\mathbf{x}_i)` and :math:`d_1(\mathbf{x}_i)` the shortest
        distance to a prototype  with  a different and the same label respectively.

        Parameters
        ----------
        model : LVQBaseClass
            The model which can be any LVQBaseClass compatible with this objective function.

        data: ndarray with shape (n_samples, n_features)
            The data.

        labels: ndarray with shape (n_samples)
            The labels of the samples in the data.

        Returns
        -------
        float:
            The cost
        """
        dist_same, dist_diff, _, _ = _compute_distance(data, labels, model)

        return np.sum(self.activation(self.discriminant(dist_same, dist_diff)))

    def gradient(
        self,
        model: "LVQBaseClass",
        data: np.ndarray,
        labels: np.ndarray,
    ) -> np.ndarray:
        r"""Computes the generalized learning objective's gradient with respect to the
        prototype with a different label:

            .. math::
                \frac{\partial E}{\partial \mathbf{w}_0} = \frac{\partial f}{\partial \mu}
                \frac{\partial \mu}{\partial d_0} \frac{\partial d_0}{\partial \mathbf{w}_0}

        with :math:`\mathbf{w}_0` the prototype with a different label than the data and :math:`d_0`
        the distance to that prototype.

            .. math::
                 \frac{\partial E}{\partial \mathbf{w}_1} = \frac{\partial f}{\partial \mu}
                 \frac{\partial \mu}{\partial d_1} \frac{\partial d_1}{\partial \mathbf{w}_1}

        with :math:`\mathbf{w}_1` the prototype with the same label as the data and :math:`d_1`
        the distance to that prototype.

        Parameters
        ----------
        model : LVQBaseClass
            The model which can be any LVQBaseClass compatible with this objective function.

        data: ndarray with shape (n_samples, n_features)
            The data.

        labels: ndarray with shape (n_samples)
            The labels of the samples in the data.

        Returns
        -------
        ndarray with the same shape as the model variables array (depending on the model)
            The generalized learning objective function's gradient

        """

        dist_same, dist_diff, i_dist_same, i_dist_diff = _compute_distance(
            data, labels, model
        )
        discriminant_score = self.discriminant(dist_same, dist_diff)

        # Pre-allocation, needs to be zero.
        gradient_buffer = np.zeros(model.get_variables().size)

        # For each prototype
        for i_prototype in range(0, model.prototypes_labels_.size):
            # Find for which samples it is the closest/winner AND has the same label
            # ii_winner_same = i_prototype == i_dist_same
            if i_prototype in i_dist_same:
                ii_winner_same = i_prototype == i_dist_same
                # Only if these cases exist we can/should compute an update
                self._partial_gradient(
                    gradient_buffer,
                    discriminant_score[ii_winner_same],
                    dist_same[ii_winner_same],
                    dist_diff[ii_winner_same],
                    True,  # Indicating same label
                    data[ii_winner_same, :],
                    model,
                    i_prototype,
                )

            # Find for which samples this prototype is the closest and has a different label
            if i_prototype in i_dist_diff:
                ii_winner_diff = i_prototype == i_dist_diff
                self._partial_gradient(
                    gradient_buffer,
                    discriminant_score[ii_winner_diff],
                    dist_same[ii_winner_diff],
                    dist_diff[ii_winner_diff],
                    False,  # Indicating diff label
                    data[ii_winner_diff, :],
                    model,
                    i_prototype,
                )

        return gradient_buffer

    def _partial_gradient(
        self,
        gradient_buffer,
        discriminant_score,
        dist_same,
        dist_diff,
        winner_same,
        data,
        model,
        i_prototype,
    ):
        # Computes the following partial derivative: df/du
        activation_gradient = self.activation.gradient(discriminant_score)

        #  Computes the following partial derivatives: du/ddi, with i = 2
        discriminant_gradient = self.discriminant.gradient(
            dist_same, dist_diff, winner_same
        )

        # Computes the following partial derivatives: ddi/dwi, with i = 2
        distance_gradient = model._distance.gradient(data, model, i_prototype)

        # The distance vectors weighted by the activation and discriminant partial
        # derivatives.
        model.add_partial_gradient(
            gradient_buffer,
            (activation_gradient * discriminant_gradient).dot(distance_gradient),
            i_prototype,
        )


def _find_min(indices: np.ndarray, distances: np.ndarray) -> (np.ndarray, np.ndarray):
    """ Helper function to find the minimum distance and the index of this distance. """
    # Set the irrelevant distances to infinity.
    dist_temp = np.where(indices, distances, np.inf)
    # Find the indices of the closest prototype (column)
    i_dist_min = dist_temp.argmin(axis=1)
    # Return the shortest distances and the indices of the prototypes.
    return dist_temp[np.arange(i_dist_min.size), i_dist_min], i_dist_min


def _compute_distance(data: np.ndarray, labels: np.ndarray, model: "LVQBaseClass"):
    """Computes the distances between each prototype and each observation and finds all indices
    where the shortest distance is that of the prototype with the same label and with a different label."""
    prototypes_labels = model.prototypes_labels_

    # Step 1: Compute distances between X and the model (how is depending on model and coupled
    # distance function)
    distances = model._distance(data, model)

    # Step 2: Find for all samples the distance between closest prototype with same label (d1)
    # and different label (d2). ii_same marks for all samples the prototype with the same label.

    num_samples = labels.size
    num_prototypes = model.prototypes_labels_.size
    if num_samples == 1:
        # Faster if num_samples == 1
        ii_same = np.atleast_2d(labels == prototypes_labels)
    elif num_samples < num_prototypes:
        # Faster to go over the labels if there are less than the prototypes.
        ii_same = np.array([label == prototypes_labels for label in labels])
    else:
        # List comprehension of the prototypes. This are all slight improvements to computation
        # time, as list comprehension takes quite some time.
        ii_same = np.transpose(
            [labels == prototype_label for prototype_label in prototypes_labels]
        )

    # For each prototype mark the samples that have a different label
    ii_diff = ~ii_same

    # For each sample find the closest prototype with the same label. Returns distance and index
    # of prototype
    dist_same, i_dist_same = _find_min(ii_same, distances)

    # For each sample find the closest prototype with a different label. Returns distance and
    # index of prototype
    dist_diff, i_dist_diff = _find_min(ii_diff, distances)

    return dist_same, dist_diff, i_dist_same, i_dist_diff



import pytest
import numpy as np

from sklearn.utils.estimator_checks import check_estimator
from sklearn import datasets

from .. import GLVQ
from .. import GMLVQ
from .. import LGMLVQ

from sklvq.activations import Identity


@pytest.mark.parametrize("estimator", [GLVQ, GMLVQ, LGMLVQ])
def test_estimators(estimator):
    instance = estimator()
    return check_estimator(instance)


@pytest.mark.parametrize("estimator", [GLVQ, GMLVQ, LGMLVQ])
def test_common_hyper_parameters(estimator):
    X, y = datasets.load_iris(return_X_y=True)

    # Prototype initialization
    with pytest.raises(ValueError):
        estimator(prototype_n_per_class=np.array([1, 1])).fit(X, y)

    with pytest.raises(ValueError):
        estimator(prototype_n_per_class="100").fit(
            np.array([[1, 2, 3], [1, 2, 3]]), np.array([1, 2])
        )

    with pytest.raises(ValueError):
        estimator(prototype_n_per_class=np.array([1, 0, 1])).fit(X, y)

    with pytest.raises(ValueError):
        estimator(prototype_init="abc").fit(X, y)

    m = estimator(prototype_n_per_class=np.array([1, 2, 1])).fit(X, y)
    assert m.prototypes_.shape[0] == 4

    # Activation string which does not exist
    with pytest.raises(ValueError):
        estimator(activation_type="abc123").fit(X, y)

    # Activation object instead of type
    activation_type = Identity()
    with pytest.raises(ValueError):
        estimator(activation_type=activation_type).fit(X, y)

    activation_type = Identity
    with pytest.raises(TypeError):
        estimator(activation_type=activation_type, activation_params={"beta": 0}).fit(
            X, y
        )


@pytest.mark.parametrize("estimator", [GMLVQ, LGMLVQ])
def test_shared_memory(estimator):
    X, y = datasets.load_iris(return_X_y=True)
    m = estimator(activation_type="identity").fit(X, y)

    p = m.prototypes_
    m.set_prototypes(np.random.random(size=(3, 4)))
    assert np.shares_memory(p, m.get_prototypes())
    assert np.all(m.get_prototypes() == m.prototypes_)

    o = m.omega_
    m.set_omega(np.random.random(size=m.omega_.shape))
    assert np.shares_memory(o, m.get_omega())
    assert np.all(m.get_omega() == m.omega_)

    model_params = m.to_model_params_view(m.get_variables())
    (p, o) = model_params
    assert np.all(m.prototypes_.shape == p.shape)
    assert np.shares_memory(m.prototypes_, m.get_variables())
    assert np.shares_memory(p, m.get_variables())
    assert np.shares_memory(o, m.get_variables())

    old_variables = m.get_variables()

    p = np.random.random(size=p.shape)
    o = np.random.random(size=o.shape)

    m.set_model_params((p, o))
    assert np.all(m.prototypes_.shape == p.shape)
    assert np.shares_memory(m.prototypes_, m.get_variables())
    assert not np.shares_memory(p, m.get_variables())
    assert not np.shares_memory(o, m.get_variables())

    assert np.shares_memory(old_variables, m.get_variables())


@pytest.mark.parametrize("estimator", [GMLVQ, LGMLVQ])
def test_shared_hyper_params(estimator):
    X, y = datasets.load_iris(return_X_y=True)

    with pytest.raises(ValueError):
        estimator(relevance_normalization="True").fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_normalization=6.0).fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_n_components="none").fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_n_components=120).fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_n_components=-120).fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_n_components=6.0).fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_init="abc").fit(X, y)

    estimator(relevance_init="random").fit(X, y)

    with pytest.raises(ValueError):
        estimator(relevance_init=6).fit(X, y)

    X_hat = estimator().fit(X, y).transform(X, scale=True)
    assert not np.all(np.isnan(X_hat))

    assert pytest.approx(X_hat, estimator().fit_transform(X, y, scale=True))



import pytest
import numpy as np

from sklearn import datasets
from sklearn import preprocessing
from sklearn.pipeline import make_pipeline

from sklvq.models import GLVQ


class ProgressLogger:
    def __init__(self, k=4):
        self.states = np.array([])
        self.counter = 0
        self.k = k

    def __call__(self, state) -> bool:
        self.states = np.append(self.states, state)

        if self.counter == self.k:
            return True

        self.counter = self.counter + 1

        return False


class ScipyProgressLogger:
    def __init__(self):
        self.states = np.array([])
        self.counter = 0

    def __call__(self, xk) -> bool:
        self.states = np.append(self.states, {"variables": xk, "nit": self.counter})
        self.counter = self.counter + 1

        if self.counter == 4:
            return True

        return False


@pytest.mark.parametrize(
    "solver",
    [
        "steepest-gradient-descent",
        "waypoint-gradient-descent",
        "adaptive-moment-estimation",
    ],
)
def test_solvers_callback(solver):
    iris = datasets.load_iris()

    for k in [0, 1, 2, 4, 10]:
        logger = ProgressLogger(k=k)

        estimator = GLVQ(solver_type=solver, solver_params={"callback": logger})

        pipeline = make_pipeline(preprocessing.StandardScaler(), estimator)

        pipeline.fit(iris.data, iris.target)

        if k == 0:
            assert logger.states[-1]["nit"] == "Initial"
        else:
            assert logger.states[-1]["nit"] == k


@pytest.mark.parametrize("solver", ["lbfgs", "bfgs"])
def test_scipy_solvers_callback(solver):
    iris = datasets.load_iris()

    logger = ScipyProgressLogger()

    estimator = GLVQ(solver_type=solver, solver_params={"callback": logger})

    pipeline = make_pipeline(preprocessing.StandardScaler(), estimator)

    pipeline.fit(iris.data, iris.target)


# def check_init_solver(solver_string):
#     distatance_class = init_class(distances, distance_string)
#
#     assert isinstance(distatance_class, type)
#
#     distance_instance = distatance_class()
#
#     assert isinstance(distance_instance, DistanceBaseClass)
#
#     return distatance_class
#
# def test_aliases():
#     for value in ALIASES.keys():
#         check_init_activation(value)



from ._base import DistanceBaseClass
from ._euclidean import Euclidean
from ._squared_euclidean import SquaredEuclidean
from ._adaptive_squared_euclidean import AdaptiveSquaredEuclidean
from ._local_adaptive_squared_euclidean import LocalAdaptiveSquaredEuclidean

from sklvq._utils import _import_from_string

__all__ = [
    "DistanceBaseClass",
    "Euclidean",
    "SquaredEuclidean",
    "AdaptiveSquaredEuclidean",
    "LocalAdaptiveSquaredEuclidean",
]

ALIASES = {}


def import_from_string(class_string, valid_strings=None) -> type:
    return _import_from_string(
        __name__, class_string, ALIASES, "distance_type", valid_strings
    )



import numpy as np
from sklearn.utils import shuffle

from . import SolverBaseClass
from ..objectives import ObjectiveBaseClass
from ._base import _update_state

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sklvq.models import LVQBaseClass

STATE_KEYS = ["variables", "nit", "fun", "m_hat", "v_hat"]


class AdaptiveMomentEstimation(SolverBaseClass):
    r"""Adaptive moment estimation (ADAM)

    Implementation and description inspired by `[1]`_.

    Adam maintains two moving averages of the gradient (:math:`m, v`), which get updated for
    every sample at each epoch/run until the maximum runs (``max_runs``) has been reached:

    .. math::
        \mathbf{m} &= \beta_1 \cdot \mathbf{m} + (1 - \beta_1) \cdot \nabla e_i(\theta) \\
        \mathbf{v} &= \beta_2 \cdot \mathbf{v} + (1 - \beta_2) \cdot [\nabla e_i(\theta)]^{\circ 2}.

    Since :math:`m`  and :math:`v` are initialized to zero vectors, they are biased towards zero.
    To counteract this, unbiased estimates :math:`\hat{m}` and :math:`\hat{v}` are computed:

    .. math::
        \hat{\mathbf{m}} &= \mathbf{m} / (1 - \beta^p_1) \\
        \hat{\mathbf{v}} &= \mathbf{v} / (1 - \beta^p_2),

    where :math:`p` is initially 0, but afterwards it's increased by 1 each time before
    selecting a  new random sample. The unbiased estimates of the average gradient are then used
    for the update step:

    .. math::
        \theta = \theta - \eta \cdot \hat{\mathbf{m}} \odot \hat{\mathbf{v}}^{\circ \frac{1}{2}},

    with :math:`\eta` the ``step_size``. Additionally,  ``beta1``, and ``beta2``,  can be chosen
    by the user.

    Note that :math:`\odot` denotes the elementwise (Hadamard) product and :math:`\mathbf{x}^{
    \circ y}` the elementwise power operation.

    Parameters
    ----------
    objective: ObjectiveBaseClass, required
        This is/should be set by the algorithm.
    max_runs: int
        Number of runs over all the X. Should be >= 1
    beta1: float
        Controls the decay rate of the moving average of the gradient. Should be < 1.0
        and > 0.
    beta2: float
        Controls the decay rate of the moving average of the squared gradient. Should be < 1.0
        and > 0.
    step_size: float
        The step size to control the learning rate.
    epsilon: float
        Small value to overcome zero division

    callback: callable
        Callable with signature callable(state). If the callable returns True the solver
        will stop (early). The state object contains the following information:

        - "variables"
            Concatenated 1D ndarray of the model's parameters
        - "nit"
            The current iteration counter
        - "fun"
            The objective cost
        - "m_hat"
            Unbiased moving average of the gradient
        - "v_hat"
            Unbiased moving average of the Hadamard squared gradient

    References
    ----------
    _`[1]` LeKander, M., Biehl, M., & De Vries, H. (2017). "Empirical evaluation of gradient
    methods for matrix learning vector quantization." 12th International Workshop on
    Self-Organizing Maps and Learning Vector Quantization, Clustering and Data
    Visualization, WSOM 2017.
    """

    def __init__(
        self,
        objective: ObjectiveBaseClass,
        max_runs: int = 10,
        beta1: float = 0.9,
        beta2: float = 0.999,
        step_size: float = 0.001,
        epsilon: float = 1e-4,
        callback: callable = None,
    ):
        super().__init__(objective)
        if max_runs <= 0:
            raise ValueError(
                "{}:  Expected max_runs to be > 0, but got max_runs = {}".format(
                    type(self).__name__, max_runs
                )
            )
        self.max_runs = max_runs

        if beta1 <= 0 or beta1 > 1.0:
            raise ValueError(
                "{}:  Expected beta1 to be > 0 and <= 1.0 but got beta1 = {}".format(
                    type(self).__name__, beta1
                )
            )
        self.beta1 = beta1

        if beta2 <= 0 or beta2 > 1.0:
            raise ValueError(
                "{}:  Expected beta1 to be > 0 and <= 1.0 but got beta2 = {}".format(
                    type(self).__name__, beta2
                )
            )
        self.beta2 = beta2

        if np.any(step_size <= 0):
            raise ValueError(
                "{}:  Expected step_size to be > 0, but got step_size = {}".format(
                    type(self).__name__, step_size
                )
            )
        self.step_size = step_size

        if epsilon < 0:
            raise ValueError(
                "{}:  Expected epsilon to be > 0, but got epsilon = {}".format(
                    type(self).__name__, epsilon
                )
            )
        self.epsilon = epsilon

        if callback is not None:
            if not callable(callback):
                raise ValueError(
                    "{}:  callback is not callable.".format(type(self).__name__)
                )
        self.callback = callback

    def solve(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        model: "LVQBaseClass",
    ):
        """
        Parameters
        ----------
        data : ndarray of shape (number of observations, number of dimensions)
        labels : ndarray of size (number of observations)
        model : LVQBaseClass
            The initial model that will be changed and holds the results at the end
        """

        # Administration
        variables = model.get_variables()
        variables_size = variables.size

        # Init/allocation of moving averages (m and v in literature)
        m = np.zeros(variables_size)
        v = np.zeros(variables_size)
        p = 0

        if self.callback is not None:
            state = _update_state(
                STATE_KEYS,
                variables=np.copy(model.get_variables()),
                nit="Initial",
                fun=self.objective(model, data, labels),
            )
            if self.callback(state):
                return

        for i_run in range(0, self.max_runs):
            # Randomize order of X
            shuffled_indices = shuffle(
                range(0, labels.size), random_state=model.random_state_
            )

            shuffled_data = data[shuffled_indices, np.newaxis, :]
            shuffled_labels = labels[shuffled_indices, np.newaxis]

            for i_sample, (sample, sample_label) in enumerate(
                zip(shuffled_data, shuffled_labels)
            ):
                # Update power
                p += 1

                objective_gradient = self.objective.gradient(
                    model, sample, sample_label
                )

                # Update biased (init 0) moving gradient averages m and v.
                m = (self.beta1 * m) + ((1 - self.beta1) * objective_gradient)

                v = (self.beta2 * v) + ((1 - self.beta2) * objective_gradient ** 2)

                # Update unbiased moving gradient averages
                m_hat = m / (1 - self.beta1 ** p)

                v_hat = v / (1 - self.beta2 ** p)

                objective_gradient = (
                    self.step_size * m_hat / (np.sqrt(v_hat) + self.epsilon)
                )

                model.set_variables(
                    np.subtract(  # returns out=objective_gradient
                        model.get_variables(),
                        objective_gradient,
                        out=objective_gradient,
                    )
                )

            if self.callback is not None:
                state = _update_state(
                    STATE_KEYS,
                    variables=np.copy(model.get_variables()),
                    nit=i_run + 1,
                    fun=self.objective(model, data, labels),
                    m_hat=m_hat,
                    v_hat=v_hat,
                )
                if self.callback(state):
                    return



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from ..models._base import LVQBaseClass


class ObjectiveBaseClass(ABC):
    """Objective base class

    Abstract class for implementing objective functions. It provides abstract methods with
    expected call signatures.
    """

    __slots__ = ()

    @abstractmethod
    def __call__(
        self,
        model: "LVQBaseClass",
        data: np.ndarray,
        labels: np.ndarray,
    ):
        """The objective function

        Parameters
        ----------
        variables: ndarray with shape depending on model parameters
            Flattened 1D array of the variables that are changed, i.e., the model parameters

        model : LVQBaseClass
            The model which can be any LVQBaseClass compatible with this objective function.

        data: ndarray with shape (n_samples, n_features)
            The X

        labels: ndarray with shape (n_samples)


        Returns
        -------
        float
            The cost


        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def gradient(
        self,
        model: "LVQBaseClass",
        data: np.ndarray,
        labels: np.ndarray,
    ):
        """The objective gradient

        Parameters
        ----------
        variables: ndarray with shape depending on model parameters
            Flattened 1D array of the variables that are changed, i.e., the model parameters

        model : LVQBaseClass
            The model which can be any LVQBaseClass compatible with this objective function.

        data: ndarray with shape (n_samples, n_features)
            The X

        labels: ndarray with shape (n_samples)

        Returns
        -------
        ndarray with shape of variables
            The gradient

        """
        raise NotImplementedError("You should implement this!")



import numpy as np

from .test_common import check_init_activation

import pytest


def test_swish():
    swish_class = check_init_activation("swish")

    assert isinstance(swish_class, type)

    swish = swish_class(beta=2)

    assert swish.beta == 2

    swish = swish_class()

    assert swish.beta == 1

    wrong_betas = [-1, 0]
    for wrong_beta in wrong_betas:
        with pytest.raises(ValueError):
            swish_class(beta=wrong_beta)

    # x = 0 should be 0
    assert swish(np.array([0])) == 0
    # x < 0 is negative and goes to 0
    assert swish(np.array([-1])) < 0

    # Swish is approximately the identity function, when x gets large enough, thus gradient is 1.
    x = np.array([4, 8, 12, 20, 50, 100, 200])
    assert swish(x) == pytest.approx(x, abs=1e-1)
    assert swish.gradient(x) == pytest.approx(1, abs=1e-1)

    swish = swish_class(beta=10)

    # x = 0 should still be 0
    assert swish(np.array([0])) == 0
    # x < 0 should still be negative and goes to 0
    assert swish(np.array([-1])) < 0

    # Swish is with higher beta less approximately the identity function, when x gets large
    # enough, thus gradient is 1.
    assert swish(np.array([4, 8, 12, 20, 50, 100, 200])) == pytest.approx(
        np.array([4, 8, 12, 20, 50, 100, 200]), abs=1e-10
    )
    assert swish.gradient(x) == pytest.approx(1, abs=1e-10)

    # Potentially raising errors when division by 0.0/inf/nans happens....



import pytest
from sklvq._utils import _import_class_from_string


def test_utils():
    _import_class_from_string("sklvq.distances", "squared-euclidean")

    # Wrong package
    with pytest.raises(ImportError):
        _import_class_from_string("sklvq.activations", "squared-euclidean")

    # Wrong class
    with pytest.raises(ImportError):
        _import_class_from_string("sklvq.distances", "sigmoid")

    # Case where package/module exists but the class inside the module
    # does not have the same name.



import numpy as np

from . import ActivationBaseClass


class Identity(ActivationBaseClass):
    """Identity function

    Class that holds the identity function and gradient.

    See also
    --------
    Sigmoid, SoftPlus, Swish
    """

    __slots__ = ()

    def __call__(self, x: np.ndarray) -> np.ndarray:
        r"""Implementation of the identity function:
            .. math::

                f(\mathbf{x}) = \mathbf{x}

        Parameters
        ----------
        x : ndarray of any shape

        Returns
        -------
        x : ndarray
            Elementwise evaluation of the identity function.
        """
        return x

    def gradient(self, x: np.ndarray) -> np.ndarray:
        r"""The identity functions's gradient:
            .. math::

                \frac{\partial f}{\partial \mathbf{x}} = \mathbf{1}

        Parameters
        ----------
        x : ndarray

        Returns
        -------
        ndarray of shape (x.shape)
            Elementwise evaluation of the identity function's gradient.
        """

        return np.ones_like(x)



import numpy as np
import pytest

from .._waypoint_gradient_descent import WaypointGradientDescent
from sklvq.objectives import GeneralizedLearningObjective


def test_steepest_gradient_descent():

    objective = GeneralizedLearningObjective(
        activation_type="identity",
        activation_params=None,
        discriminant_type="relative-distance",
        discriminant_params=None,
    )

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, max_runs=-1)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, max_runs=0)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, step_size=-1)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, step_size=np.array([1, -1]))

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, step_size=np.array([1, 0]))

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, callback=0)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, loss=-1)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, loss=1.1)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, gain=0.9)

    with pytest.raises(ValueError):
        WaypointGradientDescent(objective, k=-1)



from abc import ABC, abstractmethod
from typing import List, Any
from typing import TYPE_CHECKING

import numpy as np
import scipy.optimize as spo

from ..objectives import ObjectiveBaseClass

if TYPE_CHECKING:
    from ..models import LVQBaseClass


class SolverBaseClass(ABC):
    """Solver base class

    Abstract class for implementing solvers. Provides abstract methods with expected calls
    signatures.

    See also
    --------
    SteepestGradientDescent, WaypointGradientDescent, AdaptiveMomentEstimation,
    BroydenFletcherGoldfarbShanno, LimitedMemoryBfgs
    """

    def __init__(self, objective: ObjectiveBaseClass):
        self.objective = objective

    @abstractmethod
    def solve(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        model: "LVQBaseClass",
    ) -> None:
        """
        Solve updates the model it is given and does not return anything.

        Parameters
        ----------
        data : ndarray of shape (number of observations, number of dimensions)
            The data.
        labels : ndarray of size (number of observations)
            The labels of the samples in the data.
        model : LVQBaseClass
            The initial model that will also hold the final result
        """
        raise NotImplementedError("You should implement this!")


class ScipyBaseSolver(SolverBaseClass):
    """ScipyBaseSolver

    Class to wrap around scipy solvers.

    See also
    --------
    BroydenFletcherGoldfarbShanno, LimitedMemoryBfgs
    """

    def __init__(self, objective, method: str = "L-BFGS-B", **kwargs):
        self.method = method
        self.params = kwargs
        super().__init__(objective)

    def _objective_wrapper(self, variables, model, data, labels):
        model.set_variables(variables)
        return self.objective(model, data, labels)

    def _objective_gradient_wrapper(self, variables, model, data, labels):
        model.set_variables(variables)
        return self.objective.gradient(model, data, labels)

    def solve(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        model: "LVQBaseClass",
    ):
        """
        Solve updates the model it is given and does not return anything.

        Parameters
        ----------
        data : ndarray of shape (number of observations, number of dimensions)
            The data.
        labels : ndarray of size (number of observations)
            The labels of the samples in the data.
        model : LVQBaseClass
            The initial model that will also hold the final result
        """
        params = {"jac": self._objective_gradient_wrapper}
        if self.params is not None:
            params.update(self.params)

        result = spo.minimize(
            self._objective_wrapper,
            model.get_variables(),
            method=self.method,
            args=(model, data, labels),
            **params
        )

        # Update model
        model.set_variables(result.x)


def _update_state(state_keys: List[str], **kwargs: Any) -> dict:
    # Helper function that can be used to update state dict. The state_keys is a  list of strings
    # indicating the keys the dictionary should hold. If not provided in the kwargs they are set
    # to None.
    state = dict.fromkeys(state_keys)
    state.update(**kwargs)
    return state



from abc import ABC, abstractmethod

import numpy as np


class DiscriminantBaseClass(ABC):
    """Discriminant base class

    Abstract class for implementing discriminant functions. Provides abstract methods with
    expected call signatures.

    Custom discriminative function '__init__' should accept any parameters as key-value pairs.

    See also
    --------
    RelativeDistance
    """

    __slots__ = ()

    @abstractmethod
    def __call__(self, dist_same: np.ndarray, dist_diff: np.ndarray) -> np.ndarray:
        """
        Should implement a discriminant function

        Parameters
        ----------
        dist_same : ndarray of shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with the same label.
        dist_diff : ndarray of shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with a different label.

        Returns
        -------
        ndarray : with shape (n_samples, 1)
            Should perform a elementwise evaluation of a discriminant function.
        """
        raise NotImplementedError("You should implement this!")

    @abstractmethod
    def gradient(
        self, dist_same: np.ndarray, dist_diff: np.ndarray, same_label: bool
    ) -> np.ndarray:
        """
        Should implement the discriminant function's  gradient

        Parameters
        ----------
        dist_same : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with the same label.
        dist_diff : ndarray with shape (n_samples, 1), with n_samples >= 1
            Shortest distance of n_samples to a prototype with a different label.
        same_label : bool
            Indicates if the gradient with respect to a prototype with the same label (True) or
            with respect to a prototype with a different label (False) needs to be computed.

        Returns
        -------
        ndarray with shape (n_sampeles, 1)
            Should perform a elementwise evaluation of a discriminant function's gradient.

        """
        raise NotImplementedError("You should implement this!")



#! /usr/bin/env python
import codecs
import os

from setuptools import find_packages, setup

# get __version__ from _version.py
ver_file = os.path.join("sklvq", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "sklvq"
DESCRIPTION = "A collection of Learning Vector Quantization algorithms compatible with scikit-learn"
with codecs.open("README.rst", encoding="utf-8-sig") as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = "Rick van Veen"
MAINTAINER_EMAIL = "r.van.veen133@gmail.com"
URL = "https://github.com/rickvanveen/sklvq"
LICENSE = "The 3-Clause BSD License"
DOWNLOAD_URL = "https://github.com/rickvanveen/sklvq"
VERSION = __version__
INSTALL_REQUIRES = ["numpy>=1.14.0", "scipy>=1.1.0", "scikit-learn>=0.23.2"]
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]
EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-cov", "coverage", "pandas"],
    "docs": ["sphinx", "sphinx-gallery", "sphinx_rtd_theme", "numpydoc", "matplotlib", "pillow"],
}

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires='>=3.6'
)
